# -*- coding: utf-8 -*-
"""
Tests for the basic xml setter functions
"""
import os
import pytest

FILE_PATH = os.path.dirname(os.path.abspath(__file__))
TEST_INPXML_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/FePt_film_SSFT_LO/files/inp2.xml')


def test_xml_set_attrib_value_no_create(load_inpxml):
    """
    Basic test of the functionality of xml_set_attrib_value_no_create
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_set_attrib_value_no_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert len(eval_xpath(root, '/fleurInput/@TEST_ATT', list_return=True)) == 0

    xmltree = xml_set_attrib_value_no_create(xmltree, '/fleurInput', 'TEST_ATT', 'test')

    assert str(eval_xpath(root, '/fleurInput/@TEST_ATT')) == 'test'


def test_xml_set_attrib_value_no_create_not_str(load_inpxml):
    """
    Test of the automatic conversion to string xml_set_attrib_value_no_create
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_set_attrib_value_no_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert len(eval_xpath(root, '/fleurInput/@TEST_ATT', list_return=True)) == 0

    xmltree = xml_set_attrib_value_no_create(xmltree, '/fleurInput', 'TEST_ATT', 2145)

    assert str(eval_xpath(root, '/fleurInput/@TEST_ATT')) == '2145'


def test_xml_set_attrib_value_no_create_errors(load_inpxml):
    """
    Test of the error messages in xml_set_attrib_value_no_create
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_set_attrib_value_no_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match='Wrong value for occurrences'):
        xml_set_attrib_value_no_create(xmltree,
                                       '/fleurInput/atomSpecies/species/mtSphere',
                                       'radius',
                                       'test',
                                       occurrences=5)

    with pytest.raises(ValueError, match='Wrong length for attribute values'):
        xml_set_attrib_value_no_create(xmltree,
                                       '/fleurInput/atomSpecies/species/mtSphere',
                                       'radius', ['test', 'too_much'],
                                       occurrences=[1])

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['2.20000000', '2.20000000']


TEST_ATTRIB_RESULTS = [['test', 'test'], ['test', 'test2'], ['test', '2214'], ['test', '2.20000000'],
                       ['2.20000000', 'test']]
TEST_ATTRIBV = ['test', ['test', 'test2'], ['test', 2214], 'test', ['test']]
TEST_OCCURENCES = [None, None, None, 0, [-1]]


@pytest.mark.parametrize('attribv, expected_result,occurrences', zip(TEST_ATTRIBV, TEST_ATTRIB_RESULTS,
                                                                     TEST_OCCURENCES))
def test_xml_set_attrib_value_no_create_all(load_inpxml, attribv, expected_result, occurrences):
    """
    Test of the functionality of xml_set_attrib_value_no_create with multiple occurrences
    of the sttribute and different values for occurrences
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_set_attrib_value_no_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['2.20000000', '2.20000000']

    xmltree = xml_set_attrib_value_no_create(xmltree,
                                             '/fleurInput/atomSpecies/species/mtSphere',
                                             'radius',
                                             attribv,
                                             occurrences=occurrences)

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == expected_result


def test_xml_set_text_no_create(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_set_text_no_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert eval_xpath(root, '/fleurInput').text.strip() == ''

    xmltree = xml_set_text_no_create(xmltree, '/fleurInput', 'TEST_TEXT')

    assert eval_xpath(root, '/fleurInput').text == 'TEST_TEXT'


def test_xml_set_text_no_create_errors(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_set_text_no_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    kpoints_xpath = '/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint'

    with pytest.raises(ValueError, match='Wrong value for occurrences'):
        xml_set_text_no_create(xmltree, kpoints_xpath, 'test', occurrences=5)

    with pytest.raises(ValueError, match='Wrong length for text values'):
        xml_set_text_no_create(xmltree, kpoints_xpath, ['test', 'too_much'], occurrences=[1])

    assert eval_xpath(root, f'{kpoints_xpath}/text()') == [
        '   -0.250000     0.250000     0.000000', '    0.250000     0.250000     0.000000'
    ]


TEST_TEXT_RESULTS = [['test', 'test'], ['test', 'test2'], ['test', '    0.250000     0.250000     0.000000'],
                     ['   -0.250000     0.250000     0.000000', 'test']]
TEST_TEXTS = ['test', ['test', 'test2'], 'test', ['test']]
TEST_TEXT_OCCURENCES = [None, None, 0, [-1]]


@pytest.mark.parametrize('text, expected_result,occurrences', zip(TEST_TEXTS, TEST_TEXT_RESULTS, TEST_TEXT_OCCURENCES))
def test_xml_set_text_no_create_all(load_inpxml, text, expected_result, occurrences):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_set_text_no_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    kpoints_xpath = '/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint'

    assert eval_xpath(root, f'{kpoints_xpath}/text()') == [
        '   -0.250000     0.250000     0.000000', '    0.250000     0.250000     0.000000'
    ]

    xmltree = xml_set_text_no_create(xmltree, kpoints_xpath, text, occurrences=occurrences)

    assert eval_xpath(root, f'{kpoints_xpath}/text()') == expected_result


def test_xml_delete_tag_single(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_delete_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert len(eval_xpath(root, '/fleurInput/calculationSetup', list_return=True)) == 1

    xmltree = xml_delete_tag(xmltree, '/fleurInput/calculationSetup')

    assert len(eval_xpath(root, '/fleurInput/calculationSetup', list_return=True)) == 0


def test_xml_delete_tag_multiple(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_delete_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert len(eval_xpath(root, '/fleurInput/atomSpecies/species', list_return=True)) == 2

    xmltree = xml_delete_tag(xmltree, '/fleurInput/atomSpecies/species')

    assert len(eval_xpath(root, '/fleurInput/atomSpecies/species', list_return=True)) == 0


def test_xml_delete_att_single(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_delete_att

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    keys = set([('Kmax', '4.00000000'), ('Gmax', '10.00000000'), ('GmaxXC', '8.70000000'), ('numbands', '0')])

    node = eval_xpath(root, '/fleurInput/calculationSetup/cutoffs')

    assert set(node.attrib.items()) == keys

    xmltree = xml_delete_att(xmltree, '/fleurInput/calculationSetup/cutoffs', 'Kmax')

    node = eval_xpath(root, '/fleurInput/calculationSetup/cutoffs')

    keys.discard(('Kmax', '4.00000000'))
    assert set(node.attrib.items()) == keys


def test_xml_delete_att_multiple(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_delete_att

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['2.20000000', '2.20000000']

    xmltree = xml_delete_att(xmltree, '/fleurInput/atomSpecies/species/mtSphere', 'radius')

    assert len(eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')) == 0


def test_xml_replace_tag_single(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_replace_tag
    from lxml import etree

    replace_element = etree.Element('test_tag')
    replace_element.attrib['test_attrib'] = 'test'

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert len(eval_xpath(root, '/fleurInput/calculationSetup/cutoffs', list_return=True)) == 1

    xmltree = xml_replace_tag(xmltree, '/fleurInput/calculationSetup/cutoffs', replace_element)

    assert len(eval_xpath(root, '/fleurInput/calculationSetup/cutoffs', list_return=True)) == 0

    nodes = eval_xpath(root, '/fleurInput/calculationSetup/test_tag', list_return=True)

    assert len(nodes) == 1
    assert nodes[0].attrib.items() == [('test_attrib', 'test')]


def test_xml_replace_tag_multiple(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_replace_tag
    from lxml import etree

    replace_element = etree.Element('test_tag')
    replace_element.attrib['test_attrib'] = 'test'

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert len(eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere', list_return=True)) == 2

    xmltree = xml_replace_tag(xmltree, '/fleurInput/atomSpecies/species/mtSphere', replace_element)

    assert len(eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere', list_return=True)) == 0

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species/test_tag', list_return=True)

    assert len(nodes) == 2
    assert nodes[0].attrib.items() == [('test_attrib', 'test')]
    assert nodes[1].attrib.items() == [('test_attrib', 'test')]


def test_xml_create_tag_string_append(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    tags = [
        'cutoffs', 'scfLoop', 'coreElectrons', 'xcFunctional', 'magnetism', 'soc', 'prodBasis', 'expertModes',
        'geometryOptimization', 'ldaU'
    ]

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    assert [child.tag for child in node.iterchildren()] == tags

    xmltree = xml_create_tag(xmltree, '/fleurInput/calculationSetup', 'test_tag')

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    tags.append('test_tag')
    assert [child.tag for child in node.iterchildren()] == tags


def test_xml_create_tag_element_append(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag
    from lxml import etree

    new_element = etree.Element('test_tag')
    new_element.attrib['test_attrib'] = 'test'

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    tags = [
        'cutoffs', 'scfLoop', 'coreElectrons', 'xcFunctional', 'magnetism', 'soc', 'prodBasis', 'expertModes',
        'geometryOptimization', 'ldaU'
    ]

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    assert [child.tag for child in node.iterchildren()] == tags

    xmltree = xml_create_tag(xmltree, '/fleurInput/calculationSetup', new_element)

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    tags.append('test_tag')
    assert [child.tag for child in node.iterchildren()] == tags
    assert [child.attrib.items() for child in node.iterchildren()][-1] == [('test_attrib', 'test')]


def test_xml_create_tag_insert_first(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    tags = [
        'cutoffs', 'scfLoop', 'coreElectrons', 'xcFunctional', 'magnetism', 'soc', 'prodBasis', 'expertModes',
        'geometryOptimization', 'ldaU'
    ]

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    assert [child.tag for child in node.iterchildren()] == tags

    xmltree = xml_create_tag(xmltree, '/fleurInput/calculationSetup', 'test_tag', place_index=0)

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    assert [child.tag for child in node.iterchildren()] == [
        'test_tag', 'cutoffs', 'scfLoop', 'coreElectrons', 'xcFunctional', 'magnetism', 'soc', 'prodBasis',
        'expertModes', 'geometryOptimization', 'ldaU'
    ]


def test_xml_create_tag_insert_middle(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    tags = [
        'cutoffs', 'scfLoop', 'coreElectrons', 'xcFunctional', 'magnetism', 'soc', 'prodBasis', 'expertModes',
        'geometryOptimization', 'ldaU'
    ]

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    assert [child.tag for child in node.iterchildren()] == tags

    xmltree = xml_create_tag(xmltree, '/fleurInput/calculationSetup', 'test_tag', place_index=5)

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    assert [child.tag for child in node.iterchildren()] == [
        'cutoffs', 'scfLoop', 'coreElectrons', 'xcFunctional', 'magnetism', 'test_tag', 'soc', 'prodBasis',
        'expertModes', 'geometryOptimization', 'ldaU'
    ]


def test_xml_create_tag_tag_order_all_single(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    tags = [
        'cutoffs', 'scfLoop', 'coreElectrons', 'xcFunctional', 'magnetism', 'soc', 'prodBasis', 'expertModes',
        'geometryOptimization', 'ldaU'
    ]

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    assert [child.tag for child in node.iterchildren()] == tags

    order = [
        'cutoffs', 'scfLoop', 'coreElectrons', 'xcFunctional', 'magnetism', 'test_tag', 'soc', 'prodBasis',
        'expertModes', 'geometryOptimization', 'ldaU'
    ]
    xmltree = xml_create_tag(xmltree, '/fleurInput/calculationSetup', 'test_tag', tag_order=order)

    node = eval_xpath(root, '/fleurInput/calculationSetup')

    assert [child.tag for child in node.iterchildren()] == order


def test_xml_create_tag_tag_order_multiple(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    tags = [[
        'mtSphere',
        'atomicCutoffs',
        'electronConfig',
        'energyParameters',
        'lo',
        'lo',
    ], ['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags

    order = ['mtSphere', 'atomicCutoffs', 'electronConfig', 'test_tag', 'energyParameters', 'lo']
    xmltree = xml_create_tag(xmltree, '/fleurInput/atomSpecies/species', 'test_tag', tag_order=order)

    tags = [[
        'mtSphere',
        'atomicCutoffs',
        'electronConfig',
        'test_tag',
        'energyParameters',
        'lo',
        'lo',
    ], ['mtSphere', 'atomicCutoffs', 'electronConfig', 'test_tag', 'energyParameters', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags


def test_xml_create_tag_tag_order_multiple_selection(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    tags = [[
        'mtSphere',
        'atomicCutoffs',
        'electronConfig',
        'energyParameters',
        'lo',
        'lo',
    ], ['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags

    order = ['mtSphere', 'atomicCutoffs', 'electronConfig', 'test_tag', 'energyParameters', 'lo']
    xmltree = xml_create_tag(xmltree, "/fleurInput/atomSpecies/species[@name='Fe-1']", 'test_tag', tag_order=order)

    tags = [[
        'mtSphere',
        'atomicCutoffs',
        'electronConfig',
        'test_tag',
        'energyParameters',
        'lo',
        'lo',
    ], ['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags


def test_xml_create_tag_tag_order_multiple_beginning(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    tags = [[
        'mtSphere',
        'atomicCutoffs',
        'electronConfig',
        'energyParameters',
        'lo',
        'lo',
    ], ['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags

    order = ['test_tag', 'mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']
    xmltree = xml_create_tag(xmltree, '/fleurInput/atomSpecies/species', 'test_tag', tag_order=order)

    tags = [[
        'test_tag',
        'mtSphere',
        'atomicCutoffs',
        'electronConfig',
        'energyParameters',
        'lo',
        'lo',
    ], ['test_tag', 'mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags


def test_xml_create_tag_tag_order_multiple_occurrences_single(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    tags = [[
        'mtSphere',
        'atomicCutoffs',
        'electronConfig',
        'energyParameters',
        'lo',
        'lo',
    ], ['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags

    order = ['test_tag', 'mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']
    xmltree = xml_create_tag(xmltree, '/fleurInput/atomSpecies/species', 'test_tag', tag_order=order, occurrences=0)

    tags = [[
        'test_tag',
        'mtSphere',
        'atomicCutoffs',
        'electronConfig',
        'energyParameters',
        'lo',
        'lo',
    ], ['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags


def test_xml_create_tag_tag_order_multiple_occurrences_list(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    tags = [[
        'mtSphere',
        'atomicCutoffs',
        'electronConfig',
        'energyParameters',
        'lo',
        'lo',
    ], ['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags

    order = ['test_tag', 'mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']
    xmltree = xml_create_tag(xmltree, '/fleurInput/atomSpecies/species', 'test_tag', tag_order=order, occurrences=[-1])

    tags = [[
        'mtSphere',
        'atomicCutoffs',
        'electronConfig',
        'energyParameters',
        'lo',
        'lo',
    ], ['test_tag', 'mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags


def test_xml_create_tag_errors(load_inpxml):

    from masci_tools.util.xml.xml_setters_basic import xml_create_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)

    with pytest.raises(ValueError, match=r"Could not create tag 'test_tag' because atleast one subtag is missing."):
        xml_create_tag(xmltree, '/fleurInput/calculationSetup/not_existent', 'test_tag')

    order = ['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']
    with pytest.raises(ValueError, match=r"The tag 'test_tag' was not found in the order list"):
        xml_create_tag(xmltree, '/fleurInput/atomSpecies/species', 'test_tag', tag_order=order)

    order = ['atomicCutoffs', 'electronConfig', 'energyParameters', 'lo']
    with pytest.raises(ValueError, match=r"Did not find existing elements in the tag_order list: {'mtSphere'}"):
        xml_create_tag(xmltree, '/fleurInput/atomSpecies/species', 'lo', tag_order=order)


def test_xml_create_tag_misaligned_order(load_inpxml):
    """
    Test automatic correction of order
    """

    from masci_tools.util.xml.xml_setters_basic import xml_create_tag
    from masci_tools.util.xml.common_functions import eval_xpath

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    xml_create_tag(xmltree, '/fleurInput/atomSpecies/species', 'ldaU')  #This creates an invalid order
    xml_create_tag(xmltree, '/fleurInput/atomSpecies/species', 'lo')

    order = ['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'ldaU', 'lo']
    with pytest.raises(ValueError, match=r'Existing order does not correspond to tag_order list'):
        xml_create_tag(xmltree, '/fleurInput/atomSpecies/species', 'ldaU', tag_order=order, correct_order=False)

    with pytest.warns(UserWarning, match=r'Existing order does not correspond to tag_order list. Correcting it'):
        xml_create_tag(xmltree, '/fleurInput/atomSpecies/species', 'ldaU', tag_order=order)

    tags = [['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'ldaU', 'ldaU', 'lo', 'lo', 'lo'],
            ['mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'ldaU', 'ldaU', 'lo', 'lo']]

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species')

    assert [[child.tag for child in node.iterchildren()] for node in nodes] == tags



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains useful methods for initializing or modifying a n_mmp_mat file
for LDA+U
"""

from masci_tools.util.schema_dict_util import get_tag_xpath
import numpy as np

def set_nmmpmat(xmltree, nmmplines, schema_dict, species_name, orbital, spin,\
                state_occupations=None, orbital_occupations=None, denmat=None, phi=None, theta=None):
    """Routine sets the block in the n_mmp_mat file specified by species_name, orbital and spin
    to the desired density matrix

    :param xmltree: an xmltree that represents inp.xml
    :param nmmplines: list of lines in the n_mmp_mat file
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param species_name: string, name of the species you want to change
    :param orbital: integer, orbital quantum number of the LDA+U procedure to be modified
    :param spin: integer, specifies which spin block should be modified
    :param state_occupations: list, sets the diagonal elements of the density matrix and everything
                              else to zero
    :param denmat: matrix, specify the density matrix explicitely
    :param phi: float, optional angle (radian), by which to rotate the density matrix before writing it
    :param theta: float, optional angle (radian), by which to rotate the density matrix before writing it

    :raises ValueError: If something in the input is wrong
    :raises KeyError: If no LDA+U procedure is found on a species

    :returns: list with modified nmmplines
    """
    from masci_tools.util.xml.common_functions import eval_xpath, get_xml_attribute
    from masci_tools.util.schema_dict_util import evaluate_attribute, eval_simple_xpath, attrib_exists
    from masci_tools.io.io_nmmpmat import write_nmmpmat, write_nmmpmat_from_states, write_nmmpmat_from_orbitals

    #All lda+U procedures have to be considered since we need to keep the order
    species_base_path = get_tag_xpath(schema_dict, 'species')

    if species_name == 'all':
        species_xpath = species_base_path
    elif species_name[:4] == 'all-':  #format all-<string>
        species_xpath = f'{species_base_path}[contains(@name,"{species_name[4:]}")]'
    else:
        species_xpath = f'{species_base_path}[@name = "{species_name}"]'

    all_species = eval_xpath(xmltree, species_xpath, list_return=True)

    nspins = evaluate_attribute(xmltree, schema_dict, 'jspins')
    if 'l_mtnocoPot' in schema_dict['attrib_types']:
        if attrib_exists(xmltree, schema_dict, 'l_mtnocoPot', contains='Setup'):
            if evaluate_attribute(xmltree, schema_dict, 'l_mtnocoPot', contains='Setup'):
                nspins = 3

    if spin > nspins:
        raise ValueError(f'Invalid input: spin {spin} requested, but input has only {nspins} spins')

    all_ldau = eval_simple_xpath(xmltree, schema_dict, 'ldaU', contains='species', list_return=True)
    numRows = nspins * 14 * len(all_ldau)

    if state_occupations is not None:
        new_nmmpmat_entry = write_nmmpmat_from_states(orbital, state_occupations, phi=phi, theta=theta)
    elif orbital_occupations is not None:
        new_nmmpmat_entry = write_nmmpmat_from_orbitals(orbital, orbital_occupations, phi=phi, theta=theta)
    elif denmat is not None:
        new_nmmpmat_entry = write_nmmpmat(orbital, denmat, phi=phi, theta=theta)
    else:
        raise ValueError('Invalid definition of density matrix. Provide either state_occupations, '
                         'orbital_occupations or denmat')

    #Check that numRows matches the number of lines in nmmp_lines_copy
    #If not either there was an n_mmp_mat file present in Fleurinp before and a lda+u calculation
    #was added or removed or the n_mmp_mat file was initialized and after the fact lda+u procedures were added
    #or removed. In both cases the resolution of this modification is very involved so we throw an error
    if nmmplines is not None:
        #Remove blank lines
        while '' in nmmplines:
            nmmplines.remove('')
        if numRows != len(nmmplines):
            raise ValueError('The number of lines in n_mmp_mat does not match the number expected from '+\
                             'the inp.xml file. Either remove the existing file before making modifications '+\
                             'and only use set_nmmpmat after all modifications to the inp.xml')

    for species in all_species:
        current_name = get_xml_attribute(species, 'name')

        #Determine the place at which the given U procedure occurs
        ldau_index = None
        for index, ldau in enumerate(all_ldau):
            ldau_species = get_xml_attribute(ldau.getparent(), 'name')
            ldau_orbital = evaluate_attribute(ldau, schema_dict, 'l', contains='species')
            if current_name == ldau_species and ldau_orbital == orbital:
                ldau_index = index

        if ldau_index is None:
            raise KeyError(f'No LDA+U procedure found on species {current_name} with l={orbital}')

        #check if fleurinp has a specified n_mmp_mat file if not initialize it with 0
        if nmmplines is None:
            nmmplines = []
            for index in range(numRows):
                nmmplines.append(''.join(map(str, [f'{0.0:20.13f}' for x in range(7)])))

        #Select the right block from n_mmp_mat and overwrite it with denmatpad
        startRow = ((spin - 1) * len(all_ldau) + ldau_index) * 14

        nmmplines[startRow:startRow + 14] = new_nmmpmat_entry

    return nmmplines


def rotate_nmmpmat(xmltree, nmmplines, schema_dict, species_name, orbital, phi, theta):
    """
    Rotate the density matrix with the given angles phi and theta

    :param xmltree: an xmltree that represents inp.xml
    :param nmmplines: list of lines in the n_mmp_mat file
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param species_name: string, name of the species you want to change
    :param orbital: integer, orbital quantum number of the LDA+U procedure to be modified
    :param phi: float, angle (radian), by which to rotate the density matrix
    :param theta: float, angle (radian), by which to rotate the density matrix

    :raises ValueError: If something in the input is wrong
    :raises KeyError: If no LDA+U procedure is found on a species

    :returns: list with modified nmmplines
    """
    from masci_tools.util.xml.common_functions import eval_xpath, get_xml_attribute
    from masci_tools.util.schema_dict_util import evaluate_attribute, eval_simple_xpath, attrib_exists
    from masci_tools.io.io_nmmpmat import read_nmmpmat_block, rotate_nmmpmat_block, format_nmmpmat

    species_base_path = get_tag_xpath(schema_dict, 'species')

    if species_name == 'all':
        species_xpath = species_base_path
    elif species_name[:4] == 'all-':  #format all-<string>
        species_xpath = f'{species_base_path}[contains(@name,"{species_name[4:]}")]'
    else:
        species_xpath = f'{species_base_path}[@name = "{species_name}"]'

    all_species = eval_xpath(xmltree, species_xpath, list_return=True)

    nspins = evaluate_attribute(xmltree, schema_dict, 'jspins')
    if 'l_mtnocoPot' in schema_dict['attrib_types']:
        if attrib_exists(xmltree, schema_dict, 'l_mtnocoPot', contains='Setup'):
            if evaluate_attribute(xmltree, schema_dict, 'l_mtnocoPot', contains='Setup'):
                nspins = 3

    all_ldau = eval_simple_xpath(xmltree, schema_dict, 'ldaU', contains='species', list_return=True)
    numRows = nspins * 14 * len(all_ldau)

    #Check that numRows matches the number of lines in nmmp_lines_copy
    #If not either there was an n_mmp_mat file present in Fleurinp before and a lda+u calculation
    #was added or removed or the n_mmp_mat file was initialized and after the fact lda+u procedures were added
    #or removed. In both cases the resolution of this modification is very involved so we throw an error
    if nmmplines is not None:
        #Remove blank lines
        while '' in nmmplines:
            nmmplines.remove('')
        if numRows != len(nmmplines):
            raise ValueError('The number of lines in n_mmp_mat does not match the number expected from '+\
                             'the inp.xml file. Either remove the existing file before making modifications '+\
                             'and only use set_nmmpmat after all modifications to the inp.xml')
    else:
        raise ValueError('rotate_nmmpmat has to be called with a initialized density matrix')

    for species in all_species:
        current_name = get_xml_attribute(species, 'name')

        #Determine the place at which the given U procedure occurs
        ldau_index = None
        for index, ldau in enumerate(all_ldau):
            ldau_species = get_xml_attribute(ldau.getparent(), 'name')
            ldau_orbital = evaluate_attribute(ldau, schema_dict, 'l', contains='species')
            if current_name == ldau_species and ldau_orbital == orbital:
                ldau_index = index

        if ldau_index is None:
            raise KeyError(f'No LDA+U procedure found on species {current_name} with l={orbital}')

        denmat = []

        for spin in range(nspins):

            startRow = (spin * len(all_ldau) + ldau_index) * 14
            denmat = read_nmmpmat_block(nmmplines, spin * len(all_ldau) + ldau_index)

            denmat = rotate_nmmpmat_block(denmat, orbital, phi=phi, theta=theta)

            nmmplines[startRow:startRow + 14] = format_nmmpmat(denmat)

    return nmmplines


def validate_nmmpmat(xmltree, nmmplines, schema_dict):
    """
    Checks that the given nmmp_lines is valid with the given xmltree

    Checks that the number of blocks is as expected from the inp.xml and each
    block does not contain non-zero elements outside their size given by the
    orbital quantum number in the inp.xml. Additionally the occupations, i.e.
    diagonal elements are checked that they are in between 0 and the maximum
    possible occupation

    :param xmltree: an xmltree that represents inp.xml
    :param nmmplines: list of lines in the n_mmp_mat file

    :raises ValueError: if any of the above checks are violated.
    """
    from masci_tools.util.xml.common_functions import get_xml_attribute
    from masci_tools.util.schema_dict_util import evaluate_attribute, eval_simple_xpath, attrib_exists

    nspins = evaluate_attribute(xmltree, schema_dict, 'jspins')
    if 'l_mtnocoPot' in schema_dict['attrib_types']:
        if attrib_exists(xmltree, schema_dict, 'l_mtnocoPot', contains='Setup'):
            if evaluate_attribute(xmltree, schema_dict, 'l_mtnocoPot', contains='Setup'):
                nspins = 3

    all_ldau = eval_simple_xpath(xmltree, schema_dict, 'ldaU', contains='species', list_return=True)
    numRows = nspins * 14 * len(all_ldau)

    tol = 0.01
    if nspins > 1:
        maxOcc = 1.0
    else:
        maxOcc = 2.0

    #Check that numRows matches the number of lines in nmmp_lines
    if nmmplines is not None:
        #Remove blank lines
        while '' in nmmplines:
            nmmplines.remove('')
        if numRows != len(nmmplines):
            raise ValueError('The number of lines in n_mmp_mat does not match the number expected from '+\
                             'the inp.xml file.')
    else:
        return

    #Now check for each block if the numbers make sense
    #(no numbers outside the valid area and no nonsensical occupations)
    for ldau_index, ldau in enumerate(all_ldau):

        orbital = evaluate_attribute(ldau, schema_dict, 'l', contains='species')
        species_name = get_xml_attribute(ldau.getparent(), 'name')

        for spin in range(nspins):
            startRow = (spin * len(all_ldau) + ldau_index) * 14

            for index in range(startRow, startRow + 14):
                currentLine = index - startRow
                currentRow = currentLine // 2

                line = nmmplines[index].split('    ')
                while '' in line:
                    line.remove('')
                nmmp = np.array([float(x) for x in line])

                outside_val = False
                if abs(currentRow - 3) > orbital:
                    if any(np.abs(nmmp) > 1e-12):
                        outside_val = True

                if currentLine % 2 == 0:
                    #m=-3 to m=0 real part
                    if any(np.abs(nmmp[:(3 - orbital) * 2]) > 1e-12):
                        outside_val = True

                else:
                    #m=0 imag part to m=3
                    if any(np.abs(nmmp[orbital * 2 + 1:]) > 1e-12):
                        outside_val = True

                if outside_val:
                    raise ValueError(f'Found value outside of valid range in for species {species_name}, spin {spin+1}'
                                     f' and l={orbital}')

                invalid_diag = False
                if spin < 2:
                    if currentRow - 3 <= 0 and currentLine % 2 == 0:
                        if nmmp[currentRow * 2] < -tol or nmmp[currentRow * 2] > maxOcc + tol:
                            invalid_diag = True
                    else:
                        if nmmp[(currentRow - 3) * 2 - 1] < -tol or nmmp[(currentRow - 3) * 2 - 1] > maxOcc + tol:
                            invalid_diag = True

                if invalid_diag:
                    raise ValueError(f'Found invalid diagonal element for species {species_name}, spin {spin+1}'
                                     f' and l={orbital}')



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Common functions for converting types to and from XML files
"""


def convert_xml_attribute(stringattribute, possible_types, constants=None, logger=None, list_return=False):
    """
    Tries to converts a given string attribute to the types given in possible_types.
    First succeeded conversion will be returned

    If no logger is given and a attribute cannot be converted an error is raised

    :param stringattribute: str, Attribute to convert.
    :param possible_types: list of str What types it will try to convert to
    :param constants: dict, of constants defined in fleur input
    :param logger: logger object for logging warnings
                   if given the errors are logged and the list is returned with the unconverted values
                   otherwise a error is raised, when the first conversion fails
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :return: The converted value of the first successful conversion
    """
    from masci_tools.util.fleur_calculate_expression import calculate_expression

    if not isinstance(stringattribute, list):
        stringattribute = [stringattribute]

    exceptions = []

    converted_list = []
    all_success = True
    for attrib in stringattribute:
        for value_type in possible_types:
            if value_type == 'float':
                try:
                    converted_value = float(attrib)
                except (ValueError, TypeError) as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'float_expression':
                if constants is None:
                    raise ValueError(
                        "For calculating attributes of the type 'float_expression' constants have to be given")
                try:
                    converted_value = calculate_expression(attrib, constants)
                except ValueError as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'int':
                try:
                    converted_value = int(attrib)
                except (ValueError, TypeError) as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'switch':
                try:
                    converted_value = convert_from_fortran_bool(attrib)
                except (ValueError, TypeError) as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'string':
                converted_value = str(attrib)

            converted_list.append(converted_value)
            break
        else:
            if logger is None:
                raise ValueError(f"Could not convert '{attrib}'. Tried: {possible_types}.\n"
                                 'The following errors occurred:\n   ' + '\n   '.join([str(exc) for exc in exceptions]))
            else:
                logger.warning("Could not convert '%s'. The following errors occurred:", attrib)

                for exc in exceptions:
                    logger.warning('   %s', str(exc))
                    logger.debug(exc, exc_info=exc)

            converted_list.append(attrib)
            all_success = False

    ret_value = converted_list
    if len(converted_list) == 1 and not list_return:
        ret_value = converted_list[0]

    return ret_value, all_success


def convert_attribute_to_xml(attributevalue, possible_types, logger=None, float_format='.10', list_return=False):
    """
    Tries to converts a given attributevalue to a string for a xml file according
    to the types given in possible_types.
    First succeeded conversion will be returned

    :param attributevalue: value to convert.
    :param possible_types: list of str What types it will try to convert from
    :param logger: logger object for logging warnings
                   if given the errors are logged and the list is returned with the unconverted values
                   otherwise a error is raised, when the first conversion fails
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :return: The converted str of the value of the first succesful conversion
    """

    if not isinstance(attributevalue, list):
        attributevalue = [attributevalue]

    possible_types = possible_types.copy()

    if 'int' in possible_types:  #Since it just converts to string
        possible_types.remove('int')
        possible_types.append('int')

    if 'string' in possible_types:  #Move string to back
        possible_types.remove('string')

    possible_types.append('string')  #Always try string

    converted_list = []
    exceptions = []
    all_success = True
    for value in attributevalue:

        for value_type in possible_types:
            if value_type in ('float', 'float_expression'):
                try:
                    converted_value = f'{value:{float_format}f}'
                except ValueError as exc:
                    exceptions.append(exc)
                    continue

            elif value_type == 'switch':
                try:
                    converted_value = convert_to_fortran_bool(value)
                except (ValueError, TypeError) as exc:
                    exceptions.append(exc)
                    continue

            elif value_type in ('string', 'int'):
                converted_value = str(value)

            converted_list.append(converted_value)
            break
        else:
            if logger is None:
                raise ValueError(f"Could not convert '{value}' to text. Tried: {possible_types}.\n"
                                 'The following errors occurred:\n   ' + '\n   '.join([str(exc) for exc in exceptions]))
            else:
                logger.warning("Could not convert '%s' to text. The following errors occurred:", value)

                for exc in exceptions:
                    logger.warning('   %s', str(exc))
                    logger.debug(exc, exc_info=exc)

            converted_list.append(value)
            all_success = False

    ret_value = converted_list
    if len(converted_list) == 1 and not list_return:
        ret_value = converted_list[0]

    return ret_value, all_success


def convert_xml_text(tagtext, possible_definitions, constants=None, logger=None, list_return=False):
    """
    Tries to converts a given string text based on the definitions (length and type).
    First succeeded conversion will be returned

    :param tagtext: str, text to convert.
    :param possible_defintions: list of dicts What types it will try to convert to
    :param constants: dict, of constants defined in fleur input
    :param logger: logger object for logging warnings
                   if given the errors are logged and the list is returned with the unconverted values
                   otherwise a error is raised, when the first conversion fails
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :return: The converted value of the first succesful conversion
    """

    if not isinstance(tagtext, list):
        tagtext = [tagtext]

    converted_list = []
    all_success = True
    for text in tagtext:

        base_text = text.strip()
        split_text = text.split(' ')
        while '' in split_text:
            split_text.remove('')

        text_definition = None
        for definition in possible_definitions:
            if definition['length'] == len(split_text):
                text_definition = definition

        if text_definition is None:
            for definition in possible_definitions:
                if definition['length'] == 'unbounded' or \
                   definition['length'] == 1:
                    text_definition = definition

        if text_definition is None:
            if logger is None:
                raise ValueError(f"Failed to convert '{text}', no matching definition found")
            else:
                logger.warning("Failed to convert '%s', no matching definition found", text)
            converted_list.append(text)
            all_success = False
            continue

        #Avoid splitting the text accidentally if there is no length restriction
        if text_definition['length'] == 1:
            split_text = [base_text]

        converted_text = []
        for value in split_text:
            converted_value, suc = convert_xml_attribute(value,
                                                         text_definition['type'],
                                                         constants=constants,
                                                         logger=logger)
            converted_text.append(converted_value)
            if not suc:
                all_success = False

        if len(converted_text) == 1 and text_definition['length'] != 'unbounded':
            converted_text = converted_text[0]

        converted_list.append(converted_text)

    ret_value = converted_list
    if len(converted_list) == 1 and not list_return:
        ret_value = converted_list[0]

    return ret_value, all_success


def convert_text_to_xml(textvalue, possible_definitions, logger=None, float_format='16.13', list_return=False):
    """
    Tries to convert a given list of values to str for a xml file based on the definitions (length and type).
    First succeeded conversion will be returned

    :param textvalue: value to convert
    :param possible_definitions: list of dicts What types it will try to convert to
    :param logger: logger object for logging warnings
                   if given the errors are logged and the list is returned with the unconverted values
                   otherwise a error is raised, when the first conversion fails
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :return: The converted value of the first succesful conversion
    """

    if not isinstance(textvalue, list):
        textvalue = [textvalue]
    elif not isinstance(textvalue[0], list):
        textvalue = [textvalue]

    converted_list = []
    all_success = True
    for text in textvalue:

        if not isinstance(text, list):
            text = [text]

        text_definition = None
        for definition in possible_definitions:
            if definition['length'] == len(text):
                text_definition = definition

        if text_definition is None:
            for definition in possible_definitions:
                if definition['length'] == 'unbounded':
                    text_definition = definition

        if text_definition is None:
            if isinstance(text, str):
                converted_list.append(text)
                continue

            if logger is None:
                raise ValueError(f"Failed to convert '{text}', no matching definition found")
            else:
                logger.warning("Failed to convert '%s', no matching definition found", text)

            converted_list.append('')
            all_success = False
            continue

        converted_text = []
        for value in text:
            converted_value, suc = convert_attribute_to_xml(value,
                                                            text_definition['type'],
                                                            logger=logger,
                                                            float_format=float_format)
            converted_text.append(converted_value)
            if not suc:
                if isinstance(value, str):
                    converted_list.append(value)
                    continue
                all_success = False

        converted_list.append(' '.join(converted_text))

    ret_value = converted_list
    if len(converted_list) == 1 and not list_return:
        ret_value = converted_list[0]

    return ret_value, all_success


def convert_from_fortran_bool(stringbool):
    """
    Converts a string in this case ('T', 'F', or 't', 'f') to True or False

    :param stringbool: a string ('t', 'f', 'F', 'T')

    :return: boolean  (either True or False)
    """

    true_items = ['True', 't', 'T']
    false_items = ['False', 'f', 'F']
    if isinstance(stringbool, str):
        if stringbool in false_items:
            return False
        elif stringbool in true_items:
            return True
        else:
            raise ValueError(f"Could not convert: '{stringbool}' to boolean, "
                             "which is not 'True', 'False', 't', 'T', 'F' or 'f'")
    elif isinstance(stringbool, bool):
        return stringbool  # no conversion needed...

    raise TypeError(f"Could not convert: '{stringbool}' to boolean, " 'only accepts str or boolean')


def convert_to_fortran_bool(boolean):
    """
    Converts a Boolean as string to the format defined in the input

    :param boolean: either a boolean or a string ('True', 'False', 'F', 'T')

    :return: a string (either 't' or 'f')
    """

    if isinstance(boolean, bool):
        if boolean:
            return 'T'
        else:
            return 'F'
    elif isinstance(boolean, str):  # basestring):
        if boolean in ('True', 't', 'T'):
            return 'T'
        elif boolean in ('False', 'f', 'F'):
            return 'F'
        else:
            raise ValueError(f"A string: {boolean} for a boolean was given, which is not 'True',"
                             "'False', 't', 'T', 'F' or 'f'")

    raise TypeError('convert_to_fortran_bool accepts only a string or ' f'bool as argument, given {boolean} ')


def convert_fleur_lo(loelements):
    """
    Converts lo xml elements from the inp.xml file into a lo string for the inpgen
    """
    # Developer hint: Be careful with using '' and "", basestring and str are not the same...
    # therefore other conversion methods might fail, or the wrong format could be written.
    from masci_tools.util.xml.common_functions import get_xml_attribute

    shell_map = {0: 's', 1: 'p', 2: 'd', 3: 'f'}

    lo_string = ''
    for element in loelements:
        lo_type = get_xml_attribute(element, 'type')
        if lo_type != 'SCLO':  # non standard los not supported for now
            continue
        l_num = get_xml_attribute(element, 'l')
        n_num = get_xml_attribute(element, 'n')
        lostr = f'{n_num}{shell_map[int(l_num)]}'
        lo_string = lo_string + ' ' + lostr
    return lo_string.strip()


def convert_str_version_number(version_str):
    """
    Convert the version number as a integer for easy comparisons

    :param version_str: str of the version number, e.g. '0.33'

    :returns: tuple of ints representing the version str
    """

    version_numbers = version_str.split('.')

    if len(version_numbers) != 2:
        raise ValueError(f"Version number is malformed: '{version_str}'")

    return tuple(int(part) for part in version_numbers)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#import pytest
from masci_tools.io.modify_potential import modify_potential
from masci_tools.io.parsers.kkrimp_parser_functions import KkrimpParserFunctions


class Test_modify_potential(object):
    """ Tests for the modify_potential class functions. """

    def test_shapefun_from_scoef(self, file_regression):
        shapefun_path = '../tests/files/mod_pot/test2/shapefun'
        scoefpath = '../tests/files/mod_pot/test2/scoef'
        atom2shapes = [1]
        shapefun_new = '../tests/files/mod_pot/test2/shapefun_new'
        modify_potential().shapefun_from_scoef(scoefpath, shapefun_path, atom2shapes, shapefun_new)
        txt = open(shapefun_new).read().strip()
        file_regression.check(txt)

    def test_neworder_potential_no_replace(self, file_regression):
        path = '../tests/files/mod_pot/test1/'
        pot = path + 'pot'
        out_pot = path + 'pot_new'
        neworder = [0, 1, 2]
        # test 1: neworder_potential standard
        modify_potential().neworder_potential(pot, out_pot, neworder)
        txt = open(out_pot).read().strip()
        file_regression.check(txt)

    def test_neworder_potential_with_replace(self, file_regression):
        path = '../tests/files/mod_pot/test1/'
        pot1 = path + 'out_potential'
        pot2 = path + 'pot'
        out_pot = path + 'pot_new'
        neworder = [0, 1, 2]
        # test 2: neworder_potential with replace from second potential
        replace_newpos = [[0, 0], [2, 0]]
        modify_potential().neworder_potential(pot1, out_pot, neworder, potfile_2=pot2, replace_from_pot2=replace_newpos)
        txt = open(out_pot).read().strip()
        file_regression.check(txt)


class Test_KkrimpParserFunctions(object):
    """ Tests for the KKRimp parser functions. """

    def test_parse_outfiles_full(self, data_regression):
        path = 'files/kkrimp_parser/test1/'
        files = {}
        files['outfile'] = path + 'out_kkrimp'
        files['out_log'] = path + 'out_log.000.txt'
        files['out_pot'] = path + 'out_potential'
        files['out_enersp_at'] = path + 'out_energysp_per_atom_eV'
        files['out_enertot_at'] = path + 'out_energytotal_per_atom_eV'
        files['out_timing'] = path + 'out_timing.000.txt'
        files['kkrflex_llyfac'] = path + 'out_timing.000.txt'
        files['kkrflex_angles'] = path + 'out_timing.000.txt'
        files['out_spinmoms'] = path + 'out_magneticmoments.txt'
        files['out_orbmoms'] = path + 'out_magneticmoments.txt'
        s, m, o = KkrimpParserFunctions().parse_kkrimp_outputfile({}, files, debug=True)
        print('files:', files)
        print('\nsuccess?\n{}\n'.format(s))
        print('\nmessages?\n{}\n'.format(m))
        print('\nout_dict?\n{}\n'.format(o))

        assert s
        assert m == []
        data_regression.check(o)

    def test_parse_outfiles_full_filehandle(self, data_regression):
        path = 'files/kkrimp_parser/test1/'
        files = {}
        files['outfile'] = open(path + 'out_kkrimp')
        files['out_log'] = open(path + 'out_log.000.txt')
        files['out_pot'] = open(path + 'out_potential')
        files['out_enersp_at'] = open(path + 'out_energysp_per_atom_eV')
        files['out_enertot_at'] = open(path + 'out_energytotal_per_atom_eV')
        files['out_timing'] = open(path + 'out_timing.000.txt')
        files['kkrflex_llyfac'] = open(path + 'out_timing.000.txt')  # file not there yet and not parsed
        files['kkrflex_angles'] = open(path + 'out_timing.000.txt')  # file not there yet and not parsed
        files['out_spinmoms'] = open(path + 'out_timing.000.txt')  # file not there yet and not parsed
        files['out_orbmoms'] = open(path + 'out_timing.000.txt')  # file not there yet and not parsed
        s, m, o = KkrimpParserFunctions().parse_kkrimp_outputfile({}, files)
        print('files:', files)
        print('\nsuccess?\n{}\n'.format(s))
        print('\nmessages?\n{}\n'.format(m))
        print('\nout_dict?\n{}\n'.format(o))

        assert s
        assert m == []
        data_regression.check(o)

    def test_parse_file_errors(self):
        files = {}
        files['outfile'] = 'no_file_there'
        files['out_log'] = 'no_file_there'
        files['out_pot'] = 'no_file_there'
        files['out_enersp_at'] = 'no_file_there'
        files['out_enertot_at'] = 'no_file_there'
        files['out_timing'] = 'no_file_there'
        files['kkrflex_llyfac'] = 'no_file_there'
        files['kkrflex_angles'] = 'no_file_there'
        files['out_spinmoms'] = 'no_file_there'
        files['out_orbmoms'] = 'no_file_there'
        s, m, o = KkrimpParserFunctions().parse_kkrimp_outputfile({}, files, debug=True)
        print('\nsuccess?\n{}\n'.format(s))
        print('\nmessages?\n{}\n'.format(m))
        print('\nout_dict?\n{}\n'.format(o))
        assert not s
        assert set(m) == set([
            'Error parsing output of KKRimp: Version Info', 'Error parsing output of KKRimp: rms-error',
            'Error parsing output of KKRimp: nspin/natom', 'Error parsing output of KKRimp: total magnetic moment',
            'Error parsing output of KKRimp: spin moment per atom', 'Error parsing output of KKRimp: orbital moment',
            'Error parsing output of KKRimp: EF', 'Error parsing output of KKRimp: total energy',
            'Error parsing output of KKRimp: search for warnings', 'Error parsing output of KKRimp: timings',
            'Error parsing output of KKRimp: single particle energies', 'Error parsing output of KKRimp: charges',
            'Error parsing output of KKRimp: energy contour', 'Error parsing output of KKRimp: core_states',
            'Error parsing output of KKRimp: scfinfo'
        ])
        assert o == {'convergence_group': {}}



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Plotting routine for fleur density of states and bandstructures
"""


def plot_fleur_bands(bandsdata, bandsattributes, spinpol=True, bokeh_plot=False, weight=None, **kwargs):
    """
    Plot the bandstructure previously extracted from a `banddos.hdf` via the
    :py:class:`~masci_tools.io.parsers.hdf5.reader.HDF5Reader`

    This routine expects datasets and attributes read in with the `FleurBands`
    recipe from :py:mod:`~masci_tools.io.parsers.hdf5.recipes` or something
    producing equivalent data

    :param dosdata: dataset dict produced by the `FleurBands` recipe
    :param attributes: attributes dict produced by the `FleurBands` recipe
    :param spinpol: bool, if True (default) use the plot for spin-polarized bands if the data is spin-polarized
    :param bokeh_plot: bool (default False), if True use the bokeh routines for plotting
    :param weight: str, name of the weight (without spin suffix `_up` or `_dn`) you want to emphasize

    All other Kwargs are passed on to the underlying plot routines
        - Matplotlib: :py:func:`~masci_tools.vis.plot_methods.plot_bands()`, :py:func:`~masci_tools.vis.plot_methods.plot_spinpol_bands()`
        - Bokeh: :py:func:`~masci_tools.vis.bokeh_plots.bokeh_bands()`, :py:func:`~masci_tools.vis.bokeh_plots.bokeh_spinpol_bands()`
    """
    from masci_tools.vis.plot_methods import plot_bands, plot_spinpol_bands
    from masci_tools.vis.bokeh_plots import bokeh_bands, bokeh_spinpol_bands
    import pandas as pd

    nbands = bandsattributes['nbands']

    bandsdata = pd.DataFrame(data=bandsdata)
    special_kpoints = []
    for k_index, label in zip(bandsattributes['special_kpoint_indices'], bandsattributes['special_kpoint_labels']):
        special_kpoints.append((label, bandsdata['kpath'][(k_index * nbands) + 1]))

    if weight is not None:
        if isinstance(weight, list):
            if all(w in bandsdata for w in weight):
                if not bokeh_plot:
                    weight = [bandsdata[w] for w in weight]
            else:
                raise ValueError(f'List of weights provided but not all weights are present in bandsdata: {weight}')
        elif weight in bandsdata:
            if not bokeh_plot:
                weight = bandsdata[weight]
        if not bokeh_plot:
            if bandsattributes['spins'] == 2:
                weight = [bandsdata[f'{weight}_up'], bandsdata[f'{weight}_down']]
            else:
                weight = bandsdata[f'{weight}_up']
        else:
            if bandsattributes['spins'] == 2:
                weight = [f'{weight}_up', f'{weight}_down']
            else:
                weight = f'{weight}_up'

    plot_label = None
    if spinpol:
        plot_label = ['Spin-Up', 'Spin-Down']

    if bokeh_plot:
        if bandsattributes['spins'] == 2:
            fig = bokeh_spinpol_bands(bandsdata,
                                      weight=weight,
                                      special_kpoints=special_kpoints,
                                      legend_label=plot_label,
                                      **kwargs)
        else:
            fig = bokeh_bands(bandsdata, weight=weight, special_kpoints=special_kpoints, **kwargs)
    else:
        if bandsattributes['spins'] == 2:
            fig = plot_spinpol_bands(bandsdata['kpath'],
                                     bandsdata['eigenvalues_up'],
                                     bandsdata['eigenvalues_down'],
                                     size_data=weight,
                                     special_kpoints=special_kpoints,
                                     plot_label=plot_label,
                                     **kwargs)
        else:
            fig = plot_bands(bandsdata['kpath'],
                             bandsdata['eigenvalues_up'],
                             size_data=weight,
                             special_kpoints=special_kpoints,
                             **kwargs)

    return fig


def plot_fleur_dos(dosdata,
                   attributes,
                   spinpol=True,
                   bokeh_plot=False,
                   multiply_by_equiv_atoms=False,
                   plot_keys=None,
                   show_total=True,
                   show_interstitial=True,
                   show_sym=False,
                   show_atoms='all',
                   show_lresolved=None,
                   key_mask=None,
                   **kwargs):
    """
    Plot the density of states previously extracted from a `banddos.hdf` via the
    :py:class:`~masci_tools.io.parsers.hdf5.reader.HDF5Reader`

    This routine expects datasets and attributes read in with the `FleurDOS` (Or related DOS modes)
    recipe from :py:mod:`~masci_tools.io.parsers.hdf5.recipes` or something
    producing equivalent data

    :param dosdata: dataset dict produced by the `FleurDOS` recipe
    :param attributes: attributes dict produced by the `FleurDOS` recipe
    :param spinpol: bool, if True (default) use the plot for spin-polarized dos if the data is spin-polarized
    :param bokeh_plot: bool (default False), if True use the bokeh routines for plotting

    Arguments for selecting the DOS components to plot:
        :param plot_keys: optional str list of str, defines the labels you want to plot
        :param show_total: bool, if True (default) the total DOS is shown
        :param show_interstitial: bool, if True (default) the interstitial DOS is shown
        :param show_atoms: either 'all', None, or int or list of ints, defines, which total atom projections to show
        :param show_atoms: either 'all', None, or int or list of ints, defines, which total atom projections to show
        :param key_mask: list of bools of the same length as the number of datasets, alternative way
                         to specify, which entries to plot

    All other Kwargs are passed on to the underlying plot routines
        - Matplotlib: :py:func:`~masci_tools.vis.plot_methods.plot_dos()`, :py:func:`~masci_tools.vis.plot_methods.plot_spinpol_dos()`
        - Bokeh: :py:func:`~masci_tools.vis.bokeh_plots.bokeh_dos()`, :py:func:`~masci_tools.vis.bokeh_plots.bokeh_spinpol_dos()`
    """
    from masci_tools.vis.plot_methods import plot_dos, plot_spinpol_dos
    from masci_tools.vis.bokeh_plots import bokeh_dos, bokeh_spinpol_dos
    import pandas as pd
    import numpy as np
    from collections import Counter

    dosdata = pd.DataFrame(data=dosdata)

    if multiply_by_equiv_atoms:
        n_equiv = Counter(attributes['atoms_groups'])
        for natom in range(1, attributes['n_types'] + 1):
            for key in dosdata.keys():
                if f'MT:{natom}' in key:
                    dosdata[key] *= n_equiv[natom]

    spinpol = attributes['spins'] == 2 and spinpol and any('_down' in key for key in dosdata.keys())
    legend_labels, keys = _generate_dos_labels(dosdata, attributes, spinpol)

    if key_mask is None:
        key_mask = _select_from_Local(keys,
                                      plot_keys,
                                      spinpol,
                                      show_total=show_total,
                                      show_interstitial=show_interstitial,
                                      show_sym=show_sym,
                                      show_atoms=show_atoms,
                                      show_lresolved=show_lresolved)

    #Select the keys
    legend_labels, keys = np.array(legend_labels)[key_mask].tolist(), np.array(keys)[key_mask].tolist()

    if bokeh_plot:
        if spinpol:
            fig = bokeh_spinpol_dos(dosdata, ynames=keys, legend_label=legend_labels, **kwargs)
        else:
            fig = bokeh_dos(dosdata, ynames=keys, legend_label=legend_labels, **kwargs)
    else:
        if spinpol:
            #Remove second half of legend labels
            legend_labels[len(legend_labels) // 2:] = [None] * (len(legend_labels) // 2)

            dosdata_up = [dosdata[key].to_numpy() for key in keys if '_up' in key]
            dosdata_dn = [dosdata[key].to_numpy() for key in keys if '_down' in key]
            fig = plot_spinpol_dos(dosdata['energy_grid'], dosdata_up, dosdata_dn, plot_label=legend_labels, **kwargs)
        else:
            dosdata_up = [dosdata[key].to_numpy() for key in keys if '_up' in key]
            fig = plot_dos(dosdata['energy_grid'], dosdata_up, plot_label=legend_labels, **kwargs)

    return fig


def _dos_order(key):
    """
    Key function for sorting DOS entries in predictable order:
        1. Energy Grid
        2. General keys (Total, interstitial, ...)
        3. Atom contribution (total, orbital resolved)
    """

    if key == 'energy_grid':
        return (-1,)

    if '_up' in key:
        key = key.split('_up')[0]
        spin = 0
    else:
        key = key.split('_down')[0]
        spin = 1

    general = ('Total', 'INT', 'Sym')
    orbital_order = ('', 's', 'p', 'd', 'f')

    if key in general:
        return (spin, general.index(key))
    elif ':' in key:
        before, after = key.split(':')

        tail = after.lstrip('0123456789')
        atom_type = int(after[:-len(tail)]) if len(tail) > 0 else int(after[0])

        if tail in orbital_order:
            return (spin, len(general) + atom_type, str(orbital_order.index(tail)))
        else:
            return (spin, len(general) + atom_type, tail)

    return None


def _generate_dos_labels(dosdata, attributes, spinpol):

    labels = []
    plot_order = []
    only_spin_up = not spinpol and any('_down' in key for key in dosdata.keys())

    types_elements = []
    for itype in range(1, attributes['n_types'] + 1):
        ind = list(attributes['atoms_groups']).index(itype)
        types_elements.append(attributes['atoms_elements'][ind])

    for key in sorted(dosdata.keys(), key=_dos_order):
        if key == 'energy_grid':
            continue

        if only_spin_up and '_down' in key:
            continue

        plot_order.append(key)
        if 'INT' in key:
            key = 'Interstitial'
            if spinpol:
                key = 'Interstitial up/down'
            labels.append(key)
        elif ':' in key:  #Atom specific DOS

            before, after = key.split(':')

            tail = after.lstrip('0123456789')
            atom_type = int(after[:-len(tail)])

            atom_label = attributes['atoms_elements'][atom_type - 1]

            if types_elements.count(atom_label) != 1:
                atom_occ = types_elements[:atom_type].count(atom_label)

                atom_label = f'{atom_label}-{atom_occ}'

            if '_up' in tail:
                tail = tail.split('_up')[0]
                if spinpol:
                    tail = f'{tail} up/down'
            else:
                tail = tail.split('_down')[0]
                if spinpol:
                    tail = f'{tail} up/down'

            labels.append(f'{atom_label} {tail}')

        else:
            if '_up' in key:
                key = key.split('_up')[0]
                if spinpol:
                    key = f'{key} up/down'
            elif '_down' in key:
                key = key.split('_down')[0]
                if spinpol:
                    key = f'{key} up/down'
            labels.append(key)

    return labels, plot_order


def _select_from_Local(keys, plot_keys, spinpol, show_total, show_interstitial, show_sym, show_atoms, show_lresolved):

    #TODO: How do we do other dos modes

    if not isinstance(show_atoms, list) and show_atoms != 'all':
        if show_atoms is not None:
            show_atoms = [show_atoms]

    if not isinstance(show_lresolved, list) and show_lresolved != 'all':
        if show_lresolved is not None:
            show_lresolved = [show_lresolved]

    #initialize mask
    if spinpol:
        mask = [False] * (len(keys) // 2)
    else:
        mask = [False] * len(keys)

    mask[0] = show_total
    mask[1] = show_interstitial
    mask[2] = show_sym

    natoms = (len(mask) - 3) // 5

    if show_atoms is not None:
        for iatom in range(1, natoms + 1):
            mask[3 + (iatom - 1) * 5] = show_atoms == 'all' or iatom in show_atoms

    if show_lresolved is not None:
        for iatom in range(1, natoms + 1):
            if show_lresolved == 'all' or iatom in show_lresolved:
                mask[3 + (iatom - 1) * 5 + 1:3 + iatom * 5] = [True, True, True, True]

    if plot_keys is not None:
        if not isinstance(plot_keys, list):
            plot_keys = [plot_keys]

        for key in plot_keys:
            mask[keys.index(f'{key}_up')] = True

    if spinpol:
        mask.extend(mask)

    return mask



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module defines a small helper class to make case insensitive dictionary
lookups available naturally
"""
from masci_tools.util.lockable_containers import LockableDict
import pprint


class CaseInsensitiveDict(LockableDict):
    """
    Dict with case insensitive lookup. Used in Schema dicts to make finding
    paths for tags and attributes easier.
    Does not preserve the case of the inserted key.
    Does not support case insensitive lookups in nested dicts
    Subclass of :py:class:`masci_tools.util.lockable_containers.LockableDict`. So
    can be frozen via the`freeze()` method

    :param upper: bool if True the method `upper()` will be used instead of `lower()`
                  to normalize keys

    All other args or kwargs will be passed on to initialize the `UserDict`

    IMPORTANT NOTE:
        This is not a direct subcalss of dict. So isinstance(a, dict)
        will be False if a is an CaseInsensitiveDict

    """

    def __init__(self, *args, upper=False, **kwargs):
        self._upper = upper
        super().__init__(*args, **kwargs)

    def _norm_key(self, key):
        if self._upper:
            return key.upper()
        else:
            return key.lower()

    #Here we modify the methods needed to make the lookups case insensitive
    #Since we use UserDict these methods should be enough to modify all behaviour
    def __delitem__(self, key):
        super().__delitem__(self._norm_key(key))

    def __setitem__(self, key, value):
        super().__setitem__(self._norm_key(key), value)

    def __getitem__(self, key):
        return super().__getitem__(self._norm_key(key))

    def __contains__(self, key):
        return super().__contains__(self._norm_key(key))

    def __repr__(self):
        return f'{self.__class__.__name__}({super().__repr__()})'


class CaseInsensitiveFrozenSet(frozenset):
    """
    Frozenset (i.e. immutable set) with case insensitive membership tests. Used in Schema dicts in `tag_info`
    entries to make flexible classification easy
    Preserves the case of the entered keys (`original_case()` returns the case of the first encounter)

    :param iterable: iterable only containing str

    Note:
        There might be subtle differences to expected behaviour with the methods
        __radd__, __ror__, and so on

    """

    def __new__(cls, iterable=None):
        if iterable is not None:
            return super().__new__(cls, [key.lower() for key in iterable])
        else:
            return super().__new__(cls, [])

    def __init__(self, iterable=None):
        if iterable is not None:
            self.original_case = self._get_new_original_case(iterable)
        else:
            self.original_case = {}
        self._frozenset_iter = None  #Used for customizing the iteration behaviour
        super().__init__()

    def _get_new_original_case(self, *iterables):
        new_dict = CaseInsensitiveDict()
        for iterable in iterables:
            for key in iterable:
                if key not in new_dict:
                    if isinstance(iterable, self.__class__):
                        new_dict[key] = iterable.original_case[key]
                    else:
                        new_dict[key] = key
        new_dict.freeze()
        return new_dict

    def __contains__(self, key):
        return super().__contains__(key.lower())

    def __repr__(self):
        """Returns the repr with the orinal case of the entered keys (first encounter)"""
        if self.original_case:
            return f'{self.__class__.__name__}({set(self.original_case.values())})'
        else:
            return f'{self.__class__.__name__}()'

    def __sub__(self, other):
        return self.difference(other)

    def __and__(self, other):
        return self.intersection(other)

    def __xor__(self, other):
        return self.symmetric_difference(other)

    def __or__(self, other):
        return self.union(other)

    def __eq__(self, other):
        return super().__eq__({key.lower() for key in other})

    def __ne__(self, other):
        return super().__ne__({key.lower() for key in other})

    def __iter__(self):
        self._frozenset_iter = super().__iter__()
        return self

    def __next__(self):
        try:
            return self.original_case[next(self._frozenset_iter)]
        except StopIteration:
            self._frozenset_iter = None
            raise

    def difference(self, *others):
        new_frozenset = super().difference(*[{key.lower() for key in other} for other in others])
        new_case_dict = self._get_new_original_case(self.original_case.values(), *others)
        return self.__class__({new_case_dict[key] for key in new_frozenset})

    def symmetric_difference(self, other):
        new_frozenset = super().symmetric_difference({key.lower() for key in other})
        new_case_dict = self._get_new_original_case(self.original_case.values(), other)
        return self.__class__({new_case_dict[key] for key in new_frozenset})

    def union(self, *others):
        new_frozenset = super().union(*[{key.lower() for key in other} for other in others])
        new_case_dict = self._get_new_original_case(self.original_case.values(), *others)
        return self.__class__({new_case_dict[key] for key in new_frozenset})

    def intersection(self, *others):
        new_frozenset = super().intersection(*[{key.lower() for key in other} for other in others])
        new_case_dict = self._get_new_original_case(self.original_case.values(), *others)
        return self.__class__({new_case_dict[key] for key in new_frozenset})

    def isdisjoint(self, other):
        return super().isdisjoint({key.lower() for key in other})

    def issubset(self, other):
        return super().issubset({key.lower() for key in other})

    def issuperset(self, other):
        return super().issuperset({key.lower() for key in other})


#Define custom pretty printers for these classes
#since pprint only support UserDict subclasses if they have no custom repr


def _pprint_case_insensitive_dict(self, object, stream, indent, allowance, context, level):  #pylint: disable=redefined-builtin
    """
    Modified from pprint dict https://github.com/python/cpython/blob/3.7/Lib/pprint.py#L194
    """
    cls = object.__class__
    stream.write(cls.__name__ + '(')
    self._pprint_dict(object, stream, indent + len(cls.__name__), allowance + 1, context, level)
    stream.write(')')


pprint.PrettyPrinter._dispatch[CaseInsensitiveDict.__repr__] = _pprint_case_insensitive_dict


def _pprint_case_insensitive_frozenset(self, object, stream, indent, allowance, context, level):  #pylint: disable=redefined-builtin
    """
    Modified from pprint dict https://github.com/python/cpython/blob/3.7/Lib/pprint.py#L194
    """
    cls = object.__class__
    stream.write(cls.__name__ + '(')
    if object:
        self._pprint_set(set(object.original_case.values()), stream, indent + len(cls.__name__), allowance + 1, context,
                         level)
    stream.write(')')


pprint.PrettyPrinter._dispatch[CaseInsensitiveFrozenSet.__repr__] = _pprint_case_insensitive_frozenset



# -*- coding: utf-8 -*-
"""
@author: ruess
"""
import pytest
import numpy as np
from masci_tools.io.common_functions import (interpolate_dos, get_alat_from_bravais, search_string, angles_to_vec,
                                             vec_to_angles, get_version_info, get_corestates_from_potential,
                                             get_highest_core_state, get_ef_from_potfile, open_general,
                                             convert_to_pystd)


class Test_common_functions(object):
    """
    Tests for the common functions from tools.common_functions
    """

    def test_open_general(self):
        path = '../tests/files/kkr/kkr_run_slab_nosoc/out_kkr'
        f = open_general(path)
        l1 = len(f.readlines())
        f = open_general(f)
        l2 = len(f.readlines())
        assert l1 == l2
        assert l2 > 0

    def test_interpolate_dos(self):
        d0 = '../tests/files/interpol/complex.dos'
        ef, dos, dos_int = interpolate_dos(d0, return_original=True)
        assert ef == 0.5256
        dos_ref = np.loadtxt('../tests/files/interpol/new3.dos')
        assert (dos_int.reshape(np.shape(dos_ref)) - dos_ref).max() < 10**-4
        assert (dos == np.load('../tests/files/interpol/ref_dos.npy')).all()

    def test_interpolate_dos_filehandle(self):
        d0 = open('../tests/files/interpol/complex.dos')
        d0 = '../tests/files/interpol/complex.dos'
        ef, dos, dos_int = interpolate_dos(d0, return_original=True)
        assert ef == 0.5256
        dos_ref = np.loadtxt('../tests/files/interpol/new3.dos')
        assert (dos_int.reshape(np.shape(dos_ref)) - dos_ref).max() < 10**-4
        assert (dos == np.load('../tests/files/interpol/ref_dos.npy')).all()

    def test_get_alat_from_bravais(self):
        bravais = np.array([[0.0, 0.5, 0.5], [0.5, 0.0, 0.5], [0.5, 0.5, 0.0]])
        alat = get_alat_from_bravais(bravais)
        assert abs(alat - np.sqrt(2) / 2) < 10**-10

    def test_search_string(self):
        txt = open('files/kkr/kkr_run_dos_output/output.0.txt', 'r').readlines()
        alatline = search_string('ALAT', txt)
        noline = search_string('ALT', txt)
        assert alatline == 23
        assert noline == -1

    def test_angles_to_vec(self):
        vec = angles_to_vec(2., 45. / 180. * np.pi, 45. / 180. * np.pi)
        assert abs(vec[0] - 1.) < 10**-10
        assert abs(vec[1] - 1.) < 10**-10
        assert abs(vec[2] - np.sqrt(2)) < 10**-10
        vec = angles_to_vec(np.array([2., 3.]), np.array([45. / 180. * np.pi, np.pi]),
                            np.array([45. / 180. * np.pi, np.pi / 2]))
        assert np.sum(abs(vec - np.array([[1., 1., np.sqrt(2)], [0, 0, -3]]))) < 10**-10

    def test_vec_to_angles(self):
        m, t, p = vec_to_angles(np.array([[0, 0, 1], [1, 1, np.sqrt(2)]]))
        assert np.sum(abs(m - np.array([1, 2]))) < 10**-10
        assert np.sum(abs(t - np.array([0, np.pi / 4.]))) < 10**-10
        assert np.sum(abs(p - np.array([0, np.pi / 4.]))) < 10**-10
        m, t, p = vec_to_angles([1, 1, np.sqrt(2)])
        assert (m, t, p) == (2, np.pi / 4., np.pi / 4.)

    def test_get_version_info(self):
        version = get_version_info('files/kkr/kkr_run_dos_output/output.0.txt')
        assert version == ('v2.2-22-g4f8f5ff', 'openmp-mac', 'kkrjm_v2.2-22-g4f8f5ff_openmp-mac_20171214102522')

    def test_get_corestates_from_potential(self):
        corestates = get_corestates_from_potential('files/kkr/kkr_run_dos_output/out_potential')
        ref = ([8, 8, 8, 8], [
            np.array([
                -1866.96096949, -275.8348967, -50.32089052, -6.5316706, -248.12312965, -41.13200278, -3.832432,
                -26.5129925
            ]),
            np.array([
                -1866.96096949, -275.8348967, -50.32089052, -6.5316706, -248.12312965, -41.13200278, -3.832432,
                -26.5129925
            ]),
            np.array([
                -1866.96096949, -275.8348967, -50.32089052, -6.5316706, -248.12312965, -41.13200278, -3.832432,
                -26.5129925
            ]),
            np.array([
                -1866.96096949, -275.8348967, -50.32089052, -6.5316706, -248.12312965, -41.13200278, -3.832432,
                -26.5129925
            ])
        ], [
            np.array([0, 0, 0, 0, 1, 1, 1, 2]),
            np.array([0, 0, 0, 0, 1, 1, 1, 2]),
            np.array([0, 0, 0, 0, 1, 1, 1, 2]),
            np.array([0, 0, 0, 0, 1, 1, 1, 2])
        ])
        assert corestates[0] == ref[0]
        assert np.sum(abs(np.array(corestates[1]) - np.array(ref[1]))) < 10**-7
        assert np.sum(abs(np.array(corestates[2]) - np.array(ref[2]))) < 10**-7

    def test_get_highest_core_state(self):
        ncore = 8
        ener = np.array([
            -1866.96096949, -275.8348967, -50.32089052, -6.5316706, -248.12312965, -41.13200278, -3.832432, -26.5129925
        ])
        lval = np.array([0, 0, 0, 0, 1, 1, 1, 2])
        out = get_highest_core_state(ncore, ener, lval)
        assert out == (1, -3.832432, '4p')

    def test_get_ef_from_potfile(self):
        ef = get_ef_from_potfile('files/kkr/kkr_run_dos_output/out_potential')
        assert ef == 1.05

    def test_convert_to_pystd(self):
        test = {
            'list': [0, 1, 2],
            'nparray': np.array([0, 1, 2]),
            'nparray_conv_list': list(np.array([0, 1, 2])),
            'int': 9,
            'float': 0.9,
            'np.int': np.int64(8),
            'np.float': np.float128(9),
            'dict': {
                'list': [0, 1, 2],
                'nparray': np.array([0, 1, 2]),
                'nparray_conv_list': list(np.array([0, 1, 2])),
                'int': 9,
                'float': 0.9,
                'np.int': np.int64(8),
                'np.float': np.float128(9),
                'dict': {
                    'list': [0, 1, 2],
                    'nparray': np.array([0, 1, 2]),
                    'nparray_conv_list': list(np.array([0, 1, 2])),
                    'int': 9,
                    'float': 0.9,
                    'np.int': np.int64(8),
                    'np.float': np.float128(9)
                }
            }
        }

        # make a copy and convert the dict
        test1 = test.copy()
        test1 = convert_to_pystd(test1)

        print('original ', test)
        print('converted', test1)

        # extract datatypes for comparison
        for i in ['list', 'nparray', 'nparray_conv_list', 'int', 'float', 'np.int', 'np.float']:
            ii = test[i]
            if i == 'list':
                out0 = []
            print(i, type(ii))
            out0.append(type(ii))
            if i in ['list', 'nparray', 'nparray_conv_list']:
                for j in ii:
                    print(j, type(j))
                    out0.append(type(j))
            # converted datatypes:
            ii = test1[i]
            if i == 'list':
                out = []
            print(i, type(ii))
            out.append(type(ii))
            if i in ['list', 'nparray', 'nparray_conv_list']:
                for j in ii:
                    print(j, type(j))
                    out.append(type(j))

        # now compare datatypes:
        assert out0 == [
            list, int, int, int, np.ndarray, np.int64, np.int64, np.int64, list, int, int, int, int, float, np.int64,
            np.float128
        ]
        assert out == [list, int, int, int, list, int, int, int, list, int, int, int, int, float, int, float]


TEST_PYTHON_VALUES = [True, False, 3.14, 4, 'str', 'TEST_STRING']
TEST_QUOTE_STRINGS = [False, False, False, False, False, True]
TEST_FORTRAN_STRING = ['.true.', '.false.', '  3.1400000000d+00', '4', 'str', "'TEST_STRING'"]


@pytest.mark.parametrize('value,expected_result,quote', zip(TEST_PYTHON_VALUES, TEST_FORTRAN_STRING,
                                                            TEST_QUOTE_STRINGS))
def test_convert_to_fortran(value, expected_result, quote):
    from masci_tools.io.common_functions import convert_to_fortran

    res = convert_to_fortran(value, quote_strings=quote)

    assert res == expected_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests of the matplotib plotting functions
"""

import pytest

# prevent issue with not having a display on travis-ci
# this needs to go *before* pyplot imports
import matplotlib
matplotlib.use('Agg')
from matplotlib.pyplot import gcf
import matplotlib.pyplot as plt
#from masci_tools.io.kkr_read_shapefun_info import read_shapefun
#from masci_tools.vis.kkr_plot_shapefun import plot_shapefun
#from masci_tools.vis.kkr_plot_dos import dosplot
#from masci_tools.vis.kkr_plot_bandstruc_qdos import dispersionplot
#from masci_tools.vis.kkr_plot_FS_qdos import FSqdos2D


def test_plot_methods_imports():
    """
    Test that all expected functions are still there
    """
    from masci_tools.vis.plot_methods import set_mpl_plot_defaults
    from masci_tools.vis.plot_methods import reset_mpl_plot_defaults
    from masci_tools.vis.plot_methods import show_mpl_plot_defaults
    from masci_tools.vis.plot_methods import single_scatterplot
    from masci_tools.vis.plot_methods import multiple_scatterplots
    from masci_tools.vis.plot_methods import multi_scatter_plot
    from masci_tools.vis.plot_methods import colormesh_plot
    from masci_tools.vis.plot_methods import waterfall_plot
    from masci_tools.vis.plot_methods import surface_plot
    from masci_tools.vis.plot_methods import multiplot_moved
    from masci_tools.vis.plot_methods import multiaxis_scatterplot
    from masci_tools.vis.plot_methods import histogram
    from masci_tools.vis.plot_methods import barchart
    from masci_tools.vis.plot_methods import plot_convex_hull2d
    from masci_tools.vis.plot_methods import plot_residuen
    from masci_tools.vis.plot_methods import plot_convergence_results
    from masci_tools.vis.plot_methods import plot_convergence_results_m
    from masci_tools.vis.plot_methods import plot_lattice_constant
    from masci_tools.vis.plot_methods import plot_relaxation_results
    from masci_tools.vis.plot_methods import plot_dos
    from masci_tools.vis.plot_methods import plot_spinpol_dos
    from masci_tools.vis.plot_methods import plot_bands
    from masci_tools.vis.plot_methods import plot_spinpol_bands
    from masci_tools.vis.plot_methods import plot_one_element_corelv
    from masci_tools.vis.plot_methods import construct_corelevel_spectrum
    from masci_tools.vis.plot_methods import plot_corelevel_spectra
    #from masci_tools.vis.plot_methods import plot_fleur_bands


TEST_CHANGES = [{'markersize': 50}, {'show': False}, {'tick_paramsx': {'labelsize': 100}}]

EXPECTED_RESULT = [{
    'markersize': 50
}, {
    'show': False
}, {
    'tick_paramsx': {
        'size': 4.0,
        'width': 1.0,
        'labelsize': 100,
        'length': 5,
        'labelrotation': 0
    }
}]


@pytest.mark.parametrize('change_dict, result', zip(TEST_CHANGES, EXPECTED_RESULT))
def test_set_defaults(change_dict, result):
    """
    Test the setting of default values
    """
    from masci_tools.vis.plot_methods import plot_params
    from masci_tools.vis.plot_methods import set_mpl_plot_defaults
    from masci_tools.vis.plot_methods import reset_mpl_plot_defaults

    value_before = {}
    for key in change_dict:
        value_before[key] = plot_params[key]

    set_mpl_plot_defaults(**change_dict)

    for key, val in result.items():
        assert plot_params[key] == val

    reset_mpl_plot_defaults()

    for key, val in value_before.items():
        assert plot_params[key] == val


class TestSingleScatterPlot(object):
    """
    Test of the single_scatterplot function
    """

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/single_scatterplot/',
                                   filename='defaults.png')
    def test_default(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import single_scatterplot

        x = np.linspace(-10, 10, 100)
        y = x**2

        gcf().clear()

        single_scatterplot(x, y, xlabel='X', ylabel='Y', title='Plot Test', show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/single_scatterplot/',
                                   filename='param_change.png')
    def test_params_changed(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import single_scatterplot

        x = np.linspace(-10, 10, 100)
        y = x**2

        gcf().clear()

        single_scatterplot(x,
                           y,
                           xlabel='X',
                           ylabel='Y',
                           title='Plot Test',
                           title_fontsize=30,
                           plot_label='Test',
                           color='darkred',
                           marker='^',
                           markersize=10,
                           linestyle='',
                           show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/single_scatterplot/',
                                   filename='scale.png')
    def test_scale(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import single_scatterplot

        x = np.linspace(-10, 10, 100)
        y = x**2

        gcf().clear()

        single_scatterplot(x, y, xlabel='X', ylabel='Y', title='Plot Test', scale={'y': 'log'}, show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/single_scatterplot/',
                                   filename='limits.png')
    def test_limits(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import single_scatterplot

        x = np.linspace(-10, 10, 100)
        y = x**2

        gcf().clear()

        single_scatterplot(x,
                           y,
                           xlabel='X',
                           ylabel='Y',
                           title='Plot Test',
                           limits={
                               'y': (-100, 100),
                               'x': (0, 10)
                           },
                           show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/single_scatterplot/',
                                   filename='area.png')
    def test_area(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import single_scatterplot

        x = np.linspace(-10, 10, 100)
        y = x**2

        gcf().clear()

        single_scatterplot(x,
                           y,
                           xlabel='X',
                           ylabel='Y',
                           title='Plot Test',
                           show=False,
                           area_plot=True,
                           area_alpha=0.3,
                           marker=None,
                           color='darkblue')
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/single_scatterplot/',
                                   filename='lines.png')
    def test_lines(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import single_scatterplot

        x = np.linspace(-10, 10, 100)
        y = x**2

        gcf().clear()

        single_scatterplot(x,
                           y,
                           xlabel='X',
                           ylabel='Y',
                           title='Plot Test',
                           lines={
                               'horizontal': 50,
                               'vertical': [-5, 5]
                           },
                           show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/single_scatterplot/',
                                   filename='lines_param_change.png')
    def test_lines_param_change(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import single_scatterplot

        x = np.linspace(-10, 10, 100)
        y = x**2

        gcf().clear()

        single_scatterplot(x,
                           y,
                           xlabel='X',
                           ylabel='Y',
                           title='Plot Test',
                           lines={
                               'horizontal': 50,
                               'vertical': [-5, {
                                   'pos': 5,
                                   'color': 'darkred',
                                   'linestyle': ':',
                                   'linewidth': 10
                               }]
                           },
                           show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/single_scatterplot/',
                                   filename='limits.png')
    def test_limits_deprecated(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import single_scatterplot

        x = np.linspace(-10, 10, 100)
        y = x**2

        gcf().clear()
        with pytest.deprecated_call():
            single_scatterplot(x,
                               y,
                               xlabel='X',
                               ylabel='Y',
                               title='Plot Test',
                               limits=[(0, 10), (-100, 100)],
                               show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/single_scatterplot/',
                                   filename='scale.png')
    def test_scale_deprecated(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import single_scatterplot

        x = np.linspace(-10, 10, 100)
        y = x**2

        gcf().clear()
        with pytest.deprecated_call():
            single_scatterplot(x, y, xlabel='X', ylabel='Y', title='Plot Test', scale=[None, 'log'], show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/single_scatterplot/',
                                   filename='defaults.png')
    def test_deprecated_label(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import single_scatterplot

        x = np.linspace(-10, 10, 100)
        y = x**2

        gcf().clear()
        with pytest.deprecated_call():
            single_scatterplot(x, y, xlabel='X', ylabel='Y', title='Plot Test', plotlabel='Test', show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()


class TestMultipleScatterPlot(object):
    """
    Test of the multiple_scatterplots function
    """

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='defaults.png')
    def test_default(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()

        multiple_scatterplots(x, y, xlabel='X', ylabel='Y', title='Plot Test', show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='param_change.png')
    def test_param_change(self):
        """
        Scatterplot with a variety of changed parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()

        multiple_scatterplots(x,
                              y,
                              xlabel='X',
                              ylabel='Y',
                              title='Plot Test',
                              title_fontsize=30,
                              plot_label=['Parabola', 'Line', None, 'cosine'],
                              marker='^',
                              linewidth=[1, 3],
                              color=['darkred', 'darkblue', 'limegreen'],
                              show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='legend.png')
    def test_legend(self):
        """
        Scatterplot with setting the legend
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()

        multiple_scatterplots(x,
                              y,
                              xlabel='X',
                              ylabel='Y',
                              title='Plot Test',
                              plot_label=['Parabola', 'Line', None, 'cosine'],
                              legend=True,
                              legend_options={'fontsize': 17},
                              show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='scale_limits.png')
    def test_scale_limits(self):
        """
        Scatterplot with setting scales and limits
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()

        multiple_scatterplots(x,
                              y,
                              xlabel='X',
                              ylabel='Y',
                              title='Plot Test',
                              scale={'y': 'log'},
                              limits={
                                  'y': (0.01, 100),
                                  'x': (0, 10)
                              },
                              show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='xticks.png')
    def test_xticks(self):
        """
        Scatterplot with setting custom xticks
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()

        multiple_scatterplots(x,
                              y,
                              xlabel='X',
                              ylabel='Y',
                              title='Plot Test',
                              xticks=[-10, 3, 3, 10, 20],
                              xticklabels=[r'$\pi$', '4', 'TEST', r'$\Omega$', r'$\frac{{1}}{{4}}$'],
                              show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='dict_selection.png')
    def test_dict_selection(self):
        """
        Test the partial setting of values via integer indexed dict
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()

        multiple_scatterplots(x,
                              y,
                              xlabel='X',
                              ylabel='Y',
                              title='Plot Test',
                              marker='^',
                              color={4: 'k'},
                              plot_label={
                                  0: 'Parabola',
                                  1: 'Line',
                                  3: 'cosine'
                              },
                              linewidth={2: 5},
                              legend=True,
                              show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='area.png')
    def test_area(self):
        """
        Test the partial setting of values via integer indexed dict
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()

        multiple_scatterplots(x,
                              y,
                              xlabel='X',
                              ylabel='Y',
                              title='Plot Test',
                              marker='^',
                              color={
                                  4: 'k',
                                  3: 'darkorange'
                              },
                              area_plot={3: True},
                              plot_label={
                                  0: 'Parabola',
                                  1: 'Line',
                                  3: 'cosine'
                              },
                              linewidth={2: 5},
                              legend=True,
                              show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='scale_limits.png')
    def test_scale_limits_deprecated(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()
        with pytest.deprecated_call():
            multiple_scatterplots(x,
                                  y,
                                  xlabel='X',
                                  ylabel='Y',
                                  title='Plot Test',
                                  scale=[None, 'log'],
                                  limits=[(0, 10), (0.01, 100)],
                                  show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='xticks.png')
    def test_xticks_deprecated(self):
        """
        Scatterplot with setting custom xticks
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()
        with pytest.deprecated_call():
            multiple_scatterplots(x,
                                  y,
                                  xlabel='X',
                                  ylabel='Y',
                                  title='Plot Test',
                                  xticks=[[r'$\pi$', '4', 'TEST', r'$\Omega$', r'$\frac{{1}}{{4}}$'],
                                          [-10, 3, 3, 10, 20]],
                                  show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='legend.png')
    def test_plot_labels_deprecated(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()
        with pytest.deprecated_call():
            multiple_scatterplots(x,
                                  y,
                                  xlabel='X',
                                  ylabel='Y',
                                  title='Plot Test',
                                  plot_labels=['Parabola', 'Line', None, 'cosine'],
                                  legend=True,
                                  legend_options={'fontsize': 20},
                                  show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='legend.png')
    def test_legend_option_deprecated(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()
        with pytest.deprecated_call():
            multiple_scatterplots(x,
                                  y,
                                  xlabel='X',
                                  ylabel='Y',
                                  title='Plot Test',
                                  plot_label=['Parabola', 'Line', None, 'cosine'],
                                  legend=True,
                                  legend_option={'fontsize': 20},
                                  show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiple_scatterplots/',
                                   filename='colors.png')
    def test_colors_deprecated(self):
        """
        Scatterplot with setting colors via deprecated option
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiple_scatterplots

        x = [np.linspace(-10, 10, 100)] * 4 + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, 50 * np.sin(x[2]), 50 * np.cos(x[3]), -5 * x[4] + 30]

        gcf().clear()
        with pytest.deprecated_call():
            multiple_scatterplots(x,
                                  y,
                                  xlabel='X',
                                  ylabel='Y',
                                  title='Plot Test',
                                  colors=['darkred', 'darkblue', 'limegreen'],
                                  show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()


class TestMultiScatterPlot(object):
    """
    Test of the multi_scatter_plot function
    """

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multi_scatter_plot/',
                                   filename='defaults.png')
    def test_default(self):
        """
        Scatterplot with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multi_scatter_plot

        x = [np.linspace(-10, 10, 50)] * 2
        y = [x[0]**2, x[1] * 5 + 30]
        s = [100 * np.exp(-0.1 * x[0]**2), abs(x[1])]

        gcf().clear()

        multi_scatter_plot(x, y, size_data=s, xlabel='X', ylabel='Y', title='Plot Test', show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multi_scatter_plot/',
                                   filename='param_change.png')
    def test_param_change(self):
        """
        Scatterplot with changed parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multi_scatter_plot

        x = [np.linspace(-10, 10, 50)] * 2
        y = [x[0]**2, x[1] * 5 + 30]
        s = [100 * np.exp(-0.1 * x[0]**2), abs(x[1])]

        gcf().clear()

        multi_scatter_plot(x,
                           y,
                           size_data=s,
                           xlabel='X',
                           ylabel='Y',
                           title='Plot Test',
                           color=['darkred', 'darkorange'],
                           marker='^',
                           plot_alpha=0.6,
                           plot_label=['Parabola', 'Line'],
                           legend=True,
                           show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()


class TestMultiPlotMoved(object):
    """Test of the multiplot_moved function"""

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiplot_moved/',
                                   filename='defaults.png')
    def test_default(self):
        """
        Mulitplot_moved with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiplot_moved

        x = [np.linspace(-10, 10, 100)] * 3
        y = [x[0] * 5 + 30, 50 * np.sin(x[1]), 50 * np.cos(x[2])]

        gcf().clear()

        multiplot_moved(x, y, xlabel='X', ylabel='Y', title='Plot Test', scale_move=2.0, show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiplot_moved/',
                                   filename='param_change.png')
    def test_param_change(self):
        """
        Mulitplot_moved with changed parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiplot_moved

        x = [np.linspace(-10, 10, 100)] * 3
        y = [x[0] * 5 + 30, 50 * np.sin(x[1]), 50 * np.cos(x[2])]

        gcf().clear()

        multiplot_moved(x,
                        y,
                        xlabel='X',
                        ylabel='Y',
                        title='Plot Test',
                        plot_label=['Line', None, 'cosine'],
                        legend=True,
                        min_add=20,
                        color=['darkred', 'darkblue', 'darkorange'],
                        scale_move=2.0,
                        show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiplot_moved/', filename='area.png')
    def test_area(self):
        """
        Mulitplot_moved with changed parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiplot_moved

        x = [np.linspace(-10, 10, 100)] * 3
        y = [x[0] * 5 + 30, 50 * np.sin(x[1]), 50 * np.cos(x[2])]

        gcf().clear()

        multiplot_moved(x,
                        y,
                        xlabel='X',
                        ylabel='Y',
                        title='Plot Test',
                        area_plot={2: True},
                        area_linecolor='k',
                        show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()


class TestWaterFallPlot(object):  #pylint: disable=missing-class-docstring

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/waterfall_plot/',
                                   filename='defaults.png')
    def test_default(self):
        """
        Mulitplot_moved with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import waterfall_plot

        x = np.linspace(-1, 1, 100)
        y = np.linspace(-1, 1, 100)

        xv, yv = np.meshgrid(x, y)
        z = 10 * np.exp(-xv**2 - yv**2)
        xv, yv, z = xv.flatten(), yv.flatten(), z.flatten()

        gcf().clear()

        waterfall_plot(xv, yv, z, xlabel='X', ylabel='Y', zlabel='Z', title='Plot Test', show=False)
        # need to return the figure in order for mpl checks to work

        return gcf()


class TestSurfacePlot(object):  #pylint: disable=missing-class-docstring

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/surface_plot/', filename='defaults.png')
    def test_default(self):
        """
        Mulitplot_moved with default parameters
        """
        import numpy as np
        from masci_tools.vis.plot_methods import surface_plot

        x = np.linspace(-1, 1, 100)
        y = np.linspace(-1, 1, 100)

        xv, yv = np.meshgrid(x, y)
        z = 10 * np.exp(-xv**2 - yv**2)

        gcf().clear()

        surface_plot(xv, yv, z, xlabel='X', ylabel='Y', zlabel='Z', title='Plot Test', show=False)
        # need to return the figure in order for mpl checks to work

        return gcf()


class TestMultiAxisScatterPlot(object):  #pylint: disable=missing-class-docstring

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiaxis/', filename='defaults.png')
    def test_defaults(self):
        """
        Test of multiaxis_scatterplot with default values
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiaxis_scatterplot
        x = [np.linspace(-10, 10, 100)] * 2 + [[np.linspace(-10, 10, 100)] * 2] + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, [50 * np.sin(x[2][0]), 50 * np.cos(x[2][1])], -5 * x[3] + 30]

        gcf().clear()

        multiaxis_scatterplot(x,
                              y,
                              axes_loc=[(0, 0), (0, 1), (1, 0), (1, 1)],
                              xlabel='X',
                              ylabel='Y',
                              title=['Parabola', 'Line1', 'sin/cos', 'Line2'],
                              num_rows=2,
                              num_cols=2,
                              show=False)
        # need to return the figure in order for mpl checks to work

        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiaxis/',
                                   filename='non_standard_layout.png')
    def test_non_standard_layout(self):
        """
        Test of multiaxis_scatterplot with default values
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiaxis_scatterplot
        x = [np.linspace(-10, 10, 100)] + [[np.linspace(-10, 10, 100)] * 2] * 2
        y = [x[0]**2, [-5 * x[1][0] + 30, x[1][1] * 5 + 30], [50 * np.sin(x[2][0]), 50 * np.cos(x[2][1])]]

        gcf().clear()

        multiaxis_scatterplot(x,
                              y,
                              axes_loc=[(0, 0), (0, 1), (1, 0)],
                              axes_kwargs={1: {
                                  'rowspan': 2
                              }},
                              xlabel='X',
                              ylabel='Y',
                              title=['Parabola', 'Lines', 'sin/cos'],
                              num_rows=2,
                              num_cols=2,
                              show=False)
        # need to return the figure in order for mpl checks to work

        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiaxis/',
                                   filename='overall_param_change.png')
    def test_overall_param_change(self):
        """
        Test of multiaxis_scatterplot with a variety of parameters changed
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiaxis_scatterplot
        x = [np.linspace(-10, 10, 100)] * 2 + [[np.linspace(-10, 10, 100)] * 2] + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, [50 * np.sin(x[2][0]), 50 * np.cos(x[2][1])], -5 * x[3] + 30]

        gcf().clear()

        multiaxis_scatterplot(x,
                              y,
                              axes_loc=[(0, 0), (0, 1), (1, 0), (1, 1)],
                              xlabel='X',
                              ylabel='Y',
                              title=['Parabola', 'Line1', 'sin/cos', 'Line2'],
                              marker='^',
                              color={0: 'darkred'},
                              linewidth=10,
                              title_fontsize=30,
                              markersize=15,
                              num_rows=2,
                              num_cols=2,
                              show=False)
        # need to return the figure in order for mpl checks to work

        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/multiaxis/',
                                   filename='single_subplot_param_change.png')
    def test_single_subplot_param_change(self):
        """
        Test of multiaxis_scatterplot with a variety of parameters changed
        """
        import numpy as np
        from masci_tools.vis.plot_methods import multiaxis_scatterplot
        x = [np.linspace(-10, 10, 100)] * 2 + [[np.linspace(-10, 10, 100)] * 2] + [np.linspace(-10, 20, 100)]
        y = [x[0]**2, x[1] * 5 + 30, [50 * np.sin(x[2][0]), 50 * np.cos(x[2][1])], -5 * x[3] + 30]

        gcf().clear()

        multiaxis_scatterplot(x,
                              y,
                              axes_loc=[(0, 0), (0, 1), (1, 0), (1, 1)],
                              xlabel='X',
                              ylabel='Y',
                              title=['Parabola', 'Line1', 'sin/cos', 'Line2'],
                              subplot_params={
                                  0: {
                                      'color': 'limegreen',
                                      'scale': {
                                          'y': 'log'
                                      }
                                  },
                                  2: {
                                      'limits': {
                                          'x': (0, 10)
                                      },
                                      'color': {
                                          0: 'darkorange'
                                      },
                                      'plot_label': ['sin', 'cos'],
                                      'legend': True
                                  }
                              },
                              num_rows=2,
                              num_cols=2,
                              show=False)
        # need to return the figure in order for mpl checks to work
        return gcf()


class TestColormeshPlot(object):  #pylint: disable=missing-class-docstring

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/colormesh/', filename='defaults.png')
    def test_defaults(self):
        """
        Test of colormesh plot with default values
        """
        import numpy as np
        from masci_tools.vis.plot_methods import colormesh_plot

        x = np.linspace(0, np.pi, 100)
        y = np.linspace(0, np.pi, 100)
        x, y = np.meshgrid(x, y)
        data = np.sin(x + y)

        gcf().clear()

        colormesh_plot(x, y, data, xlabel='X', ylabel='Y', title='sin', show=False)

        # need to return the figure in order for mpl checks to work
        return gcf()


class TestHistogramPlot(object):  #pylint: disable=missing-class-docstring

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/histogram/', filename='defaults.png')
    def test_defaults(self):
        """
        Test of histogram plot with default values
        """
        import numpy as np
        from masci_tools.vis.plot_methods import histogram

        np.random.seed(19680801)
        N_points = 10000

        # Generate a normal distribution, center at x=0 and y=5
        x = np.random.randn(N_points)

        gcf().clear()

        histogram(x, show=False)

        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/histogram/', filename='param_change.png')
    def test_param_change(self):
        """
        Test of histogram plot with default values
        """
        import numpy as np
        from masci_tools.vis.plot_methods import histogram

        np.random.seed(19680801)
        N_points = 10000

        # Generate a normal distribution, center at x=0 and y=5
        x = np.random.randn(N_points)

        gcf().clear()

        histogram(x,
                  color='darkred',
                  linewidth=2,
                  plot_alpha=0.3,
                  plot_label='Normal',
                  density=True,
                  legend=True,
                  orientation='horizontal',
                  log=True,
                  show=False)

        # need to return the figure in order for mpl checks to work
        return gcf()


class TestBarchartPlot(object):  #pylint: disable=missing-class-docstring

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/barchart/', filename='defaults.png')
    def test_defaults(self):
        """
        Test of barchart plot with default values
        """
        import numpy as np
        from masci_tools.vis.plot_methods import barchart

        x = [np.linspace(0, 10, 11)] * 2
        y = [x[0]**2, [50] * 11]
        gcf().clear()

        barchart(x, y, show=False)

        # need to return the figure in order for mpl checks to work
        return gcf()


class TestResiduenPlot(object):  #pylint: disable=missing-class-docstring

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/residuen/', filename='defaults.png')
    def test_defaults(self):
        """
        Test of residuen plot with default values
        """
        import numpy as np
        from masci_tools.vis.plot_methods import plot_residuen

        np.random.seed(19680801)
        N_points = 100

        # Generate a normal distribution, center at x=0 and y=5
        rand = np.random.randn(N_points)

        x = np.linspace(-10, 10, N_points)
        fit = x**2
        real = fit + rand

        gcf().clear()

        plot_residuen(x, fit, real, show=False)

        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/residuen/', filename='no_hist.png')
    def test_no_hist(self):
        """
        Test of residuen plot without histogram
        """
        import numpy as np
        from masci_tools.vis.plot_methods import plot_residuen

        np.random.seed(19680801)
        N_points = 100

        # Generate a normal distribution, center at x=0 and y=5
        rand = np.random.randn(N_points)

        x = np.linspace(-10, 10, N_points)
        fit = x**2
        real = fit + rand

        gcf().clear()

        plot_residuen(x, fit, real, hist=False, show=False)

        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/residuen/',
                                   filename='param_change_residue.png')
    def test_param_change_residue_plot(self):
        """
        Test of residuen plot without histogram
        """
        import numpy as np
        from masci_tools.vis.plot_methods import plot_residuen

        np.random.seed(19680801)
        N_points = 100

        # Generate a normal distribution, center at x=0 and y=5
        rand = np.random.randn(N_points)

        x = np.linspace(-10, 10, N_points)
        fit = x**2
        real = fit + rand

        gcf().clear()

        plot_residuen(x,
                      fit,
                      real,
                      show=False,
                      marker='^',
                      color='darkblue',
                      xlabel='X',
                      ylabel='Test Label',
                      labelfontsize=30)

        # need to return the figure in order for mpl checks to work
        return gcf()

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/residuen/',
                                   filename='param_change_hist.png')
    def test_param_change_hist_plot(self):
        """
        Test of residuen plot without histogram
        """
        import numpy as np
        from masci_tools.vis.plot_methods import plot_residuen

        np.random.seed(19680801)
        N_points = 100

        # Generate a normal distribution, center at x=0 and y=5
        rand = np.random.randn(N_points)

        x = np.linspace(-10, 10, N_points)
        fit = x**2
        real = fit + rand

        gcf().clear()

        plot_residuen(x,
                      fit,
                      real,
                      show=False,
                      hist_kwargs={
                          'color': 'darkblue',
                          'xlabel': 'X',
                          'ylabel': 'Test Label',
                          'labelfontsize': 30,
                          'plot_label': 'Residue',
                          'legend': True
                      })

        # need to return the figure in order for mpl checks to work
        return gcf()


class TestPlotConvergenceResults(object):  #pylint: disable=missing-class-docstring

    energies = [
        -69269.46134019217, -69269.42108466873, -69269.35509388152, -69269.62486438647, -69269.51102655893,
        -69269.48862754989, -69269.48874847183, -69269.48459145911, -69269.47327003669, -69269.47248623992,
        -69269.47244891679, -69269.47645687914, -69269.47922946361, -69269.4793222245, -69269.47901836311,
        -69269.47895198638, -69269.47886053707, -69269.47875692157, -69269.47890881824, -69269.47887586526
    ]

    distances = [
        11.6508412231, 10.5637525546, 7.1938351319, 2.6117836621, 2.4735288205, 2.9455389405, 1.8364080301,
        1.4740568937, 1.8542068593, 0.9186745766, 0.900191025, 0.5290019787, 0.0979035892, 0.1098240811, 0.0717916768,
        0.0258508395, 0.0300810883, 0.0067904499, 0.0085097364, 0.0073435947
    ]

    iteration = range(len(distances))

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/convergence/', filename='defaults.png')
    def test_defaults(self):
        """
        Test of convergence plot with default values
        """
        from masci_tools.vis.plot_methods import plot_convergence_results

        gcf().clear()

        #plot_convergence produces two figures, for testing we merge them into one
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 6))

        plot_convergence_results(self.iteration, self.distances, self.energies, show=False, axis1=ax1, axis2=ax2)

        # need to return the figure in order for mpl checks to work
        return fig

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/convergence/',
                                   filename='param_change.png')
    def test_param_change(self):
        """
        Test of convergence plot with changed parameters
        """
        from masci_tools.vis.plot_methods import plot_convergence_results

        gcf().clear()

        #plot_convergence produces two figures, for testing we merge them into one
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 6))

        plot_convergence_results(self.iteration,
                                 self.distances,
                                 self.energies,
                                 show=False,
                                 axis1=ax1,
                                 axis2=ax2,
                                 linestyle='--',
                                 color='darkred',
                                 marker='s',
                                 linewidth=10,
                                 title_fontsize=20)

        # need to return the figure in order for mpl checks to work
        return fig


class TestPlotConvergenceMulti(object):  #pylint: disable=missing-class-docstring

    @pytest.mark.mpl_image_compare(baseline_dir='files/plot_methods/matplotlib/convergence_multi/',
                                   filename='defaults.png')
    def test_defaults(self):
        """
        Test of multiple convergence plot with default values
        """
        from masci_tools.vis.plot_methods import plot_convergence_results_m
        import numpy as np

        np.random.seed(19680801)
        number_iterations = np.random.randint(15, high=50, size=15)
        iteration = [np.array(range(iters)) for iters in number_iterations]

        noise_arr = [0.1 * np.random.randn(iters) + 1.0 for iters in number_iterations]

        distance_decay = np.random.rand(15)
        distance_offset = 100 * np.random.rand(15)

        energy_decay = np.random.rand(15)
        energy_offset = 20000 + 500 * np.random.rand(15)
        energy_offset2 = 1000 * np.random.rand(15)

        distances = [
            noise * offset * np.exp(-decay * iters)
            for iters, noise, decay, offset in zip(iteration, noise_arr, distance_decay, distance_offset)
        ]
        energies = [
            noise * offset2 * np.exp(-decay * iters) + offset for iters, noise, decay, offset, offset2 in zip(
                iteration, noise_arr, energy_decay, energy_offset, energy_offset2)
        ]

        gcf().clear()

        #plot_convergence produces two figures, for testing we merge them into one
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 6))

        plot_convergence_results_m(iteration, distances, energies, show=False, axis1=ax1, axis2=ax2, modes=[])

        # need to return the figure in order for mpl checks to work
        return fig



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Load both the outxml_parser and inpxml_parser
"""

from .fleur_inpxml_parser import inpxml_parser
from .fleur_outxml_parser import outxml_parser
from .task_migrations import *
from .outxml_conversions import *

__all__ = ['inpxml_parser', 'outxml_parser']



# -*- coding: utf-8 -*-
"""
Regression tests for the HDF5Reader class
"""
from masci_tools.io.common_functions import convert_to_pystd
import os

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
HDFTEST_DIR = os.path.join(CURRENT_DIR, 'files/hdf5_reader')


def test_hdf5_reader_bands(data_regression):
    """
    Tests of the bands recipe
    """
    from masci_tools.io.parsers.hdf5 import HDF5Reader
    from masci_tools.io.parsers.hdf5.recipes import FleurBands

    TEST_BANDDOS_FILE = os.path.join(HDFTEST_DIR, 'banddos_bands.hdf')

    with HDF5Reader(TEST_BANDDOS_FILE) as reader:
        data, attrs = reader.read(recipe=FleurBands)

    data_regression.check({'datasets': convert_to_pystd(data), 'attributes': convert_to_pystd(attrs)})


def test_hdf5_reader_dos(data_regression):
    """
    Tests of the dos recipe (also pass opened file handle)
    """
    from masci_tools.io.parsers.hdf5 import HDF5Reader
    from masci_tools.io.parsers.hdf5.recipes import FleurDOS

    TEST_BANDDOS_FILE = os.path.join(HDFTEST_DIR, 'banddos_dos.hdf')

    with open(TEST_BANDDOS_FILE, 'rb') as file:
        with HDF5Reader(file) as reader:
            data, attrs = reader.read(recipe=FleurDOS)

    data_regression.check({'datasets': convert_to_pystd(data), 'attributes': convert_to_pystd(attrs)})


def test_hdf5_reader_spinpol_dos(data_regression):
    """
    Tests of the dos recipe (also pass opened file handle)
    """
    from masci_tools.io.parsers.hdf5 import HDF5Reader
    from masci_tools.io.parsers.hdf5.recipes import FleurDOS

    TEST_BANDDOS_FILE = os.path.join(HDFTEST_DIR, 'banddos_spinpol_dos.hdf')

    with HDF5Reader(TEST_BANDDOS_FILE) as reader:
        data, attrs = reader.read(recipe=FleurDOS)

    data_regression.check({'datasets': convert_to_pystd(data), 'attributes': convert_to_pystd(attrs)})


def test_hdf5_reader_jdos(data_regression):
    """
    Tests of the dos recipe (also pass opened file handle)
    """
    from masci_tools.io.parsers.hdf5 import HDF5Reader
    from masci_tools.io.parsers.hdf5.recipes import FleurJDOS

    TEST_BANDDOS_FILE = os.path.join(HDFTEST_DIR, 'banddos_spinpol_dos.hdf')

    with HDF5Reader(TEST_BANDDOS_FILE) as reader:
        data, attrs = reader.read(recipe=FleurJDOS)

    data_regression.check({'datasets': convert_to_pystd(data), 'attributes': convert_to_pystd(attrs)})



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""Convenience tools only related to Python Standard Library."""

import copy
import typing
from dataclasses import field
import json
import re


def now():
    """Get now time localized to UTC."""
    from datetime import datetime
    import pytz
    return datetime.now(tz=pytz.UTC)
    # same as: pytz.UTC.localize(datetime.now())
    # not same as: datetime.now(tz=timezone.utc)


def random_string(length: int, ascii_uppercase: bool = True, ascii_lowercase: bool = True,
                  digits: bool = True, whitespace: bool = False) -> str:
    """Generate a random string of length.

    :param length: random string length.
    :param ascii_uppercase: True: include ASCII uppercase letters.
    :param ascii_lowercase: True: include ASCII lowercase letters.
    :param digits: True: include digits.
    :param whitespace: True: include whitespaces `\t\n\r\x0b\x0c`.
    :return: random string of length.
    """
    import string
    import random

    character_sets = {string.ascii_uppercase: ascii_uppercase,
                      string.ascii_lowercase: ascii_lowercase,
                      string.digits: digits,
                      string.whitespace: whitespace}
    population = ''.join([characters for characters, included in character_sets.items() if included])
    return ''.join(random.choices(population=population, k=length))


def enforce_minimum_python_version(minimum_version: tuple = (3, 7)):
    """For some operation requiring a minimum version, enforce for interpreter.

    :param minimum_version:
    :return: 0 if satisfied, sys.exit if not
    """
    import sys
    current_version = sys.version_info
    if current_version[0] == minimum_version[0] and current_version[1] >= minimum_version[1]:
        pass
    else:
        sys.stderr.write('[%s] - Error: Your Python interpreter must be %d.%d or greater (within major version %d)\n' %
                         (sys.argv[0], minimum_version[0], minimum_version[1], minimum_version[0]))
        sys.exit(-1)
    return 0


def flatten_list_of_lists(list_of_lists: list):
    return [item for sublist in list_of_lists for item in sublist]


def split_list(lst: list, n: int):
    """Yield n-sized chunks/batches from lst. Last chunk has remainder size.

    :param lst: list
    :param n: chunk size
    :return: split list (copy)
    :rtype: generator

    DEVNOTES:
    - equivalent list comprehension: [lst[i:i + n] for i in range(0, len(lst), n)]
    - as soon as a function contains 'yield', it always returns a generator. so cannot add a switch 'as_list'.
      But can just wrap the call inside list().
    """
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def sort_lists_by_list(lists: typing.List[typing.List], key_list_index: int = 0, desc: bool = False):
    """Sort multiple lists by one list.

    The key_list gets sorted. The other lists get sorted in the same order.

    Reference: https://stackoverflow.com/a/15611016/8116031

    :param lists: list of lists to be sorted
    :param key_list_index: index of list in lists to sort by
    :param desc: False: ascending order
    :return: sorted lists
    """
    # note: two-liner equivalent:
    # line1: convert (zip(a,b,c,...)) lists into list of tuples, and
    #        sort elements of this list by key_list elements, i.e. key_list-th
    #        element in each tuple.
    # line2: unpack tuple list back into orignal, sorted, lists using
    #        nested list comprehensions and tuple unpacking.
    # sorted_lists = sorted(zip(*lists), reverse=desc, key=lambda x: x[key_list])
    # return [[x[i] for x in sorted_lists] for i in range(len(lists))]

    return zip(*sorted(zip(*lists), reverse=desc, key=lambda x: x[key_list_index]))


class NoIndent(object):
    """ Value wrapper. """

    def __init__(self, value):
        self.value = value


class JSONEncoderTailoredIndent(json.JSONEncoder):
    FORMAT_SPEC = '@@{}@@'
    regex = re.compile(FORMAT_SPEC.format(r'(\d+)'))

    def __init__(self, **kwargs):
        """JSONEncoder which allows to not indent items wrapped in 'NoIndent()'.

        Reference: https://stackoverflow.com/a/13252112/8116031

        :param kwargs:
        :type kwargs:

        >>> from masci_tools.util.python_util import JSONEncoderTailoredIndent, NoIndent
        >>> import json
        >>> from string import ascii_lowercase as letters
        >>> data_structure = {
        >>>     'layer1': {
        >>>         'layer2': {
        >>>             'layer3_1': NoIndent([{"x":1,"y":7}, {"x":0,"y":4}, {"x":5,"y":3},
        >>>                                   {"x":6,"y":9},
        >>>                                   {k: v for v, k in enumerate(letters)}]),
        >>>             'layer3_2': 'string',
        >>>             'layer3_3': NoIndent([{"x":2,"y":8,"z":3}, {"x":1,"y":5,"z":4},
        >>>                                   {"x":6,"y":9,"z":8}]),
        >>>             'layer3_4': NoIndent(list(range(20))),
        >>>         }
        >>>     }
        >>> }
        >>> print(json.dumps(data_structure, cls=JSONEncoderTailoredIndent, sort_keys=True, indent=2))
        """
        # Save copy of any keyword argument values needed for use here.
        self.__sort_keys = kwargs.get('sort_keys', None)
        super(JSONEncoderTailoredIndent, self).__init__(**kwargs)

    def default(self, obj):
        return (self.FORMAT_SPEC.format(id(obj)) if isinstance(obj, NoIndent) else super(
            JSONEncoderTailoredIndent, self).default(obj))

    def encode(self, obj):
        from _ctypes import PyObj_FromPtr

        format_spec = self.FORMAT_SPEC  # Local var to expedite access.
        json_repr = super(JSONEncoderTailoredIndent, self).encode(obj)  # Default JSON.

        # Replace any marked-up object ids in the JSON repr with the
        # value returned from the json.dumps() of the corresponding
        # wrapped Python object.
        for match in self.regex.finditer(json_repr):
            # see https://stackoverflow.com/a/15012814/355230
            id = int(match.group(1))
            no_indent = PyObj_FromPtr(id)
            json_obj_repr = json.dumps(no_indent.value, sort_keys=self.__sort_keys)

            # Replace the matched id string with json formatted representation
            # of the corresponding Python object.
            json_repr = json_repr.replace('"{}"'.format(format_spec.format(id)), json_obj_repr)

        return json_repr


def dataclass_default_field(obj: typing.Any, deepcopy=True) -> typing.Any:
    """Abbreviator for defining mutable types as default values for fields of dataclasses.
    References:
    - https://stackoverflow.com/a/53870411/8116031
    - https://docs.python.org/3/library/dataclasses.html#mutable-default-values

    Use case for this: the default value is some class instance. If you just write,
    in dataclass X, y : Y = dc.field(default=Y()), all x, x2, ... instances of X's
    variables x.y will be the same object / instance of Y. Whereas with this method, all
    x.y, x2.y, ... will be different objects / instances of Y (if deepcopy). Apart from that,
    this method allows you to also define a non-default instance of Y (ie, non-empty constructor)
    like Y(foo='bar').

    :param obj: complex / mutable default value
    :param deepcopy: default True: perform deepcopy, False: perform shallow copy.
    :return: dataclass field default value

    >>> import dataclasses as dc
    >>> from typing import Dict, Tuple
    >>> @dc.dataclass
    >>> class C:
    >>>     complex_attribute: Dict[str, Tuple[int, str]] = \
    >>>     dataclass_default_field({"a": (1, "x"), "b": (1, "y")})
    """
    if deepcopy:
        return field(default_factory=lambda: copy.deepcopy(obj))
    else:
        return field(default_factory=lambda: copy.copy(obj))


class SizeEstimator:

    def __init__(self):
        """Container for various python in-memory object size estimation methods.

        Estimating Python object size is not straightforward.

        There are many solutions available online. This is just a small referenced collection of them.
        """
        import sys, humanfriendly
        self.sys = sys
        self.hf = humanfriendly

        # Custom objects know their class.
        # Function objects seem to know way too much, including modules.
        # Exclude modules as well.
        from types import ModuleType, FunctionType
        self.BLACKLIST = type, ModuleType, FunctionType

    def sizeof_simple(self, object) -> tuple:
        """For simple PSL data structures. Reference: https://stackoverflow.com/a/19865746/8116031

        :return: tuple of bytesize, humanfriendly bytesize
        """
        size = self.sys.getsizeof(object)
        return size, self.hf.format_size(size)

    def sizeof_via_blacklist(self, object):
        """sum size of object & members. Reference: https://stackoverflow.com/a/30316760/8116031."""
        from gc import get_referents

        if isinstance(object, self.BLACKLIST):
            raise TypeError('getsize() does not take argument of type: ' + str(type(object)))
        seen_ids = set()
        size = 0
        objects = [object]
        while objects:
            need_referents = []
            for object in objects:
                if not isinstance(object, self.BLACKLIST) and id(object) not in seen_ids:
                    seen_ids.add(id(object))
                    size += self.sys.getsizeof(object)
                    need_referents.append(object)
            objects = get_referents(*need_referents)

        return size, self.hf.format_size(size)

    def sizeof_via_whitelist(self, object):
        """Recursively iterate to sum size of object & members.

        'gives much more fine-grained control over the types we're going to count for memory usage,
        but has the danger of leaving types out'

        Reference: https://stackoverflow.com/a/30316760/8116031.
        """
        from numbers import Number
        from collections import Set, Mapping, deque

        zero_depth_bases = (str, bytes, Number, range, bytearray)
        _seen_ids = set()

        def inner(obj):
            obj_id = id(obj)
            if obj_id in _seen_ids:
                return 0
            _seen_ids.add(obj_id)
            size = self.sys.getsizeof(obj)
            if isinstance(obj, zero_depth_bases):
                pass  # bypass remaining control flow and return
            elif isinstance(obj, (tuple, list, Set, deque)):
                size += sum(inner(i) for i in obj)
            elif isinstance(obj, Mapping) or hasattr(obj, 'items'):
                size += sum(inner(k) + inner(v) for k, v in getattr(obj, 'items')())
            # Check for custom object instances - may subclass above too
            if hasattr(obj, '__dict__'):
                size += inner(vars(obj))
            if hasattr(obj, '__slots__'):  # can have __slots__ with __dict__
                size += sum(inner(getattr(obj, s)) for s in obj.__slots__ if hasattr(obj, s))
            return size

        size = inner(object)
        return size, self.hf.format_size(size)
